import sys

from typing import Optional, Dict, Any, Tuple
import os
import os.path as osp
import numpy as np
import pandas as pd
import zipfile
import requests
from clint.textui import progress


from tgb.linkproppred.negative_sampler import NegativeEdgeSampler
from tgb.linkproppred.tkg_negative_sampler import TKGNegativeEdgeSampler
from tgb.linkproppred.thg_negative_sampler import THGNegativeEdgeSampler
from tgb.utils.info import (
    PROJ_DIR, 
    DATA_URL_DICT, 
    DATA_VERSION_DICT, 
    DATA_EVAL_METRIC_DICT, 
    DATA_NS_STRATEGY_DICT,
    BColors
)
from tgb.utils.pre_process import (
    csv_to_pd_data,
    process_node_feat,
    process_node_type,
    csv_to_pd_data_sc,
    csv_to_pd_data_rc,
    load_edgelist_wiki,
    csv_to_tkg_data,
    csv_to_thg_data,
    csv_to_forum_data,
    csv_to_wikidata,
    csv_to_staticdata,
)
from tgb.utils.utils import save_pkl, load_pkl
from tgb.utils.utils import add_inverse_quadruples


class LinkPropPredDataset(object):
    def __init__(
        self,
        name: str,
        root: Optional[str] = "datasets",
        meta_dict: Optional[dict] = None,
        preprocess: Optional[bool] = True,
    ):
        r"""Dataset class for link prediction dataset. Stores meta information about each dataset such as evaluation metrics etc.
        also automatically pre-processes the dataset.
        Args:
            name: name of the dataset
            root: root directory to store the dataset folder
            meta_dict: dictionary containing meta information about the dataset, should contain key 'dir_name' which is the name of the dataset folder
            preprocess: whether to pre-process the dataset
        """
        self.name = name  ## original name
        # check if dataset url exist
        if self.name in DATA_URL_DICT:
            self.url = DATA_URL_DICT[self.name]
        else:
            self.url = None
            print(f"Dataset {self.name} url not found, download not supported yet.")

        
        # check if the evaluatioin metric are specified
        if self.name in DATA_EVAL_METRIC_DICT:
            self.metric = DATA_EVAL_METRIC_DICT[self.name]
        else:
            self.metric = None
            print(
                f"Dataset {self.name} default evaluation metric not found, it is not supported yet."
            )


        root = PROJ_DIR + root

        if meta_dict is None:
            self.dir_name = "_".join(name.split("-"))  ## replace hyphen with underline
            meta_dict = {"dir_name": self.dir_name}
        else:
            self.dir_name = meta_dict["dir_name"]
        self.root = osp.join(root, self.dir_name)
        self.meta_dict = meta_dict
        if "fname" not in self.meta_dict:
            self.meta_dict["fname"] = self.root + "/" + self.name + "_edgelist.csv"
            self.meta_dict["nodefile"] = None

        if name == "tgbl-flight":
            self.meta_dict["nodefile"] = self.root + "/" + "airport_node_feat.csv"

        if name == "tkgl-wikidata" or name == "tkgl-smallpedia":
            self.meta_dict["staticfile"] = self.root + "/" + self.name + "_static_edgelist.csv"
        
        if "thg" in name:
            self.meta_dict["nodeTypeFile"] = self.root + "/" + self.name + "_nodetype.csv"
        else:
            self.meta_dict["nodeTypeFile"] = None
        
        self.meta_dict["val_ns"] = self.root + "/" + self.name + "_val_ns.pkl"
        self.meta_dict["test_ns"] = self.root + "/" + self.name + "_test_ns.pkl"

        #! version check
        self.version_passed = True
        self._version_check()

        # initialize
        self._node_feat = None
        self._edge_feat = None
        self._full_data = None
        self._train_data = None
        self._val_data = None
        self._test_data = None

        # for tkg and thg
        self._edge_type = None

        #tkgl-wikidata and tkgl-smallpedia only
        self._static_data = None

        # for thg only
        self._node_type = None
        self._node_id = None

        self.download()
        # check if the root directory exists, if not create it
        if osp.isdir(self.root):
            print("Dataset directory is ", self.root)
        else:
            # os.makedirs(self.root)
            raise FileNotFoundError(f"Directory not found at {self.root}")

        if preprocess:
            self.pre_process()

        self.min_dst_idx, self.max_dst_idx = int(self._full_data["destinations"].min()), int(self._full_data["destinations"].max())

        if ('tkg' in self.name):
            if self.name in DATA_NS_STRATEGY_DICT:
                self.ns_sampler = TKGNegativeEdgeSampler(
                    dataset_name=self.name,
                    first_dst_id=self.min_dst_idx,
                    last_dst_id=self.max_dst_idx,
                    strategy=DATA_NS_STRATEGY_DICT[self.name],
                    partial_path=self.root + "/" + self.name,
                )
            else:
                raise ValueError(f"Dataset {self.name} negative sampling strategy not found.")
        elif ('thg' in self.name):
            #* need to find the smallest node id of all nodes (regardless of types)
            
            min_node_idx = min(int(self._full_data["sources"].min()), int(self._full_data["destinations"].min()))
            max_node_idx = max(int(self._full_data["sources"].max()), int(self._full_data["destinations"].max()))
            self.ns_sampler = THGNegativeEdgeSampler(
                dataset_name=self.name,
                first_node_id=min_node_idx,
                last_node_id=max_node_idx,
                node_type=self._node_type,
            )
        else:
            self.ns_sampler = NegativeEdgeSampler(
                dataset_name=self.name,
                first_dst_id=self.min_dst_idx,
                last_dst_id=self.max_dst_idx,
            )


    def _version_check(self) -> None:
        r"""Implement Version checks for dataset files
        updates the file names based on the current version number
        prompt the user to download the new version via self.version_passed variable
        """
        if (self.name in DATA_VERSION_DICT):
            version = DATA_VERSION_DICT[self.name]
        else:
            print(f"Dataset {self.name} version number not found.")
            self.version_passed = False
            return None
        
        if (version > 1):
            #* check if current version is outdated
            self.meta_dict["fname"] = self.root + "/" + self.name + "_edgelist_v" + str(int(version)) + ".csv"
            self.meta_dict["nodefile"] = None
            if self.name == "tgbl-flight":
                self.meta_dict["nodefile"] = self.root + "/" + "airport_node_feat_v" + str(int(version)) + ".csv"
            self.meta_dict["val_ns"] = self.root + "/" + self.name + "_val_ns_v" + str(int(version)) + ".pkl"
            self.meta_dict["test_ns"] = self.root + "/" + self.name + "_test_ns_v" + str(int(version)) + ".pkl"
            
            if (not osp.exists(self.meta_dict["fname"])):
                print(f"Dataset {self.name} version {int(version)} not found.")
                print(f"Please download the latest version of the dataset.")
                self.version_passed = False
                return None
        

    def download(self):
        """
        downloads this dataset from url
        check if files are already downloaded
        """
        # check if the file already exists
        if osp.exists(self.meta_dict["fname"]):
            print("raw file found, skipping download")
            return

        inp = input(
            "Will you download the dataset(s) now? (y/N)\n"
        ).lower()  # ask if the user wants to download the dataset

        if inp == "y":
            print(
                f"{BColors.WARNING}Download started, this might take a while . . . {BColors.ENDC}"
            )
            print(f"Dataset title: {self.name}")

            if self.url is None:
                raise Exception("Dataset url not found, download not supported yet.")
            else:
                r = requests.get(self.url, stream=True)
                # download_dir = self.root + "/" + "download"
                if osp.isdir(self.root):
                    print("Dataset directory is ", self.root)
                else:
                    os.makedirs(self.root)

                path_download = self.root + "/" + self.name + ".zip"
                with open(path_download, "wb") as f:
                    total_length = int(r.headers.get("content-length"))
                    for chunk in progress.bar(
                        r.iter_content(chunk_size=1024),
                        expected_size=(total_length / 1024) + 1,
                    ):
                        if chunk:
                            f.write(chunk)
                            f.flush()
                # for unzipping the file
                with zipfile.ZipFile(path_download, "r") as zip_ref:
                    zip_ref.extractall(self.root)
                print(f"{BColors.OKGREEN}Download completed {BColors.ENDC}")
                self.version_passed = True
        else:
            raise Exception(
                BColors.FAIL + "Data not found error, download " + self.name + " failed"
            )

    def generate_processed_files(self) -> pd.DataFrame:
        r"""
        turns raw data .csv file into a pandas data frame, stored on disc if not already
        Returns:
            df: pandas data frame
        """
        node_feat = None
        if not osp.exists(self.meta_dict["fname"]):
            raise FileNotFoundError(f"File not found at {self.meta_dict['fname']}")

        if self.meta_dict["nodefile"] is not None:
            if not osp.exists(self.meta_dict["nodefile"]):
                raise FileNotFoundError(
                    f"File not found at {self.meta_dict['nodefile']}"
                )
        #* for thg must have nodetypes 
        if self.meta_dict["nodeTypeFile"] is not None:
            if not osp.exists(self.meta_dict["nodeTypeFile"]):
                raise FileNotFoundError(
                    f"File not found at {self.meta_dict['nodeTypeFile']}"
                )


        OUT_DF = self.root + "/" + "ml_{}.pkl".format(self.name)
        OUT_EDGE_FEAT = self.root + "/" + "ml_{}.pkl".format(self.name + "_edge")
        OUT_NODE_ID = self.root + "/" + "ml_{}.pkl".format(self.name + "_nodeid")
        if self.meta_dict["nodefile"] is not None:
            OUT_NODE_FEAT = self.root + "/" + "ml_{}.pkl".format(self.name + "_node")
        if self.meta_dict["nodeTypeFile"] is not None:
            OUT_NODE_TYPE = self.root + "/" + "ml_{}.pkl".format(self.name + "_nodeType")

        if (osp.exists(OUT_DF)) and (self.version_passed is True):
            print("loading processed file")
            df = pd.read_pickle(OUT_DF)
            edge_feat = load_pkl(OUT_EDGE_FEAT)
            if (self.name == "tkgl-wikidata") or (self.name == "tkgl-smallpedia"):
                node_id = load_pkl(OUT_NODE_ID)
                self._node_id = node_id
            if self.meta_dict["nodefile"] is not None:
                node_feat = load_pkl(OUT_NODE_FEAT)
            if self.meta_dict["nodeTypeFile"] is not None:
                node_type = load_pkl(OUT_NODE_TYPE)
                self._node_type = node_type

        else:
            print("file not processed, generating processed file")
            if self.name == "tgbl-flight":
                df, edge_feat, node_ids = csv_to_pd_data(self.meta_dict["fname"])
            elif self.name == "tgbl-coin":
                df, edge_feat, node_ids = csv_to_pd_data_sc(self.meta_dict["fname"])
            elif self.name == "tgbl-comment":
                df, edge_feat, node_ids = csv_to_pd_data_rc(self.meta_dict["fname"])
            elif self.name == "tgbl-review":
                df, edge_feat, node_ids = csv_to_pd_data_sc(self.meta_dict["fname"])
            elif self.name == "tgbl-wiki":
                df, edge_feat, node_ids = load_edgelist_wiki(self.meta_dict["fname"])
            elif self.name == "tgbl-subreddit":
                df, edge_feat, node_ids = load_edgelist_wiki(self.meta_dict["fname"])
            elif self.name == "tgbl-uci":
                df, edge_feat, node_ids = load_edgelist_wiki(self.meta_dict["fname"])
            elif self.name == "tgbl-enron":
                df, edge_feat, node_ids = load_edgelist_wiki(self.meta_dict["fname"])
            elif self.name == "tgbl-lastfm":
                df, edge_feat, node_ids = load_edgelist_wiki(self.meta_dict["fname"])
            elif self.name == "tkgl-polecat":
                df, edge_feat, node_ids = csv_to_tkg_data(self.meta_dict["fname"])
            elif self.name == "tkgl-icews":
                df, edge_feat, node_ids = csv_to_tkg_data(self.meta_dict["fname"])
            elif self.name == "tkgl-yago":
                df, edge_feat, node_ids = csv_to_tkg_data(self.meta_dict["fname"])
            elif self.name == "tkgl-wikidata":
                df, edge_feat, node_ids = csv_to_wikidata(self.meta_dict["fname"])
                save_pkl(node_ids, OUT_NODE_ID)
                self._node_id = node_ids
            elif self.name == "tkgl-smallpedia":
                df, edge_feat, node_ids = csv_to_wikidata(self.meta_dict["fname"])
                save_pkl(node_ids, OUT_NODE_ID)
                self._node_id = node_ids
            elif self.name == "thgl-myket":
                df, edge_feat, node_ids = csv_to_thg_data(self.meta_dict["fname"])
            elif self.name == "thgl-github":
                df, edge_feat, node_ids = csv_to_thg_data(self.meta_dict["fname"])
            elif self.name == "thgl-forum":
                df, edge_feat, node_ids = csv_to_forum_data(self.meta_dict["fname"])
            elif self.name == "thgl-software":
                df, edge_feat, node_ids = csv_to_thg_data(self.meta_dict["fname"])
            else:
                raise ValueError(f"Dataset {self.name} not found.")

            save_pkl(edge_feat, OUT_EDGE_FEAT)
            df.to_pickle(OUT_DF)
            if self.meta_dict["nodefile"] is not None:
                node_feat = process_node_feat(self.meta_dict["nodefile"], node_ids)
                save_pkl(node_feat, OUT_NODE_FEAT)
            if self.meta_dict["nodeTypeFile"] is not None:
                node_type = process_node_type(self.meta_dict["nodeTypeFile"], node_ids)
                save_pkl(node_type, OUT_NODE_TYPE)
                #? do not return node_type, simply set it
                self._node_type = node_type
            

        return df, edge_feat, node_feat

    def pre_process(self):
        """
        Pre-process the dataset and generates the splits, must be run before dataset properties can be accessed
        generates the edge data and different train, val, test splits
        """

        # check if path to file is valid
        df, edge_feat, node_feat = self.generate_processed_files()

        #* design choice, only stores the original edges not the inverse relations on disc
        if ("tkgl" in self.name):
            df = add_inverse_quadruples(df)

        sources = np.array(df["u"])
        destinations = np.array(df["i"])
        timestamps = np.array(df["ts"])
        edge_idxs = np.array(df["idx"])
        weights = np.array(df["w"])
        edge_label = np.ones(len(df))  # should be 1 for all pos edges
        if (self.name == "tgbl-coin") or (self.name == "tgbl-review"):
            self._edge_feat = weights.reshape(-1,1)
        elif (self.name == "tgbl-comment"):
            self._edge_feat = np.concatenate((edge_feat, weights.reshape(-1,1)), axis=1)
        else:
            self._edge_feat = edge_feat
        self._node_feat = node_feat

        full_data = {
            "sources": sources.astype(int),
            "destinations": destinations.astype(int),
            "timestamps": timestamps.astype(int),
            "edge_idxs": edge_idxs,
            "edge_feat": self._edge_feat,
            "w": weights,
            "edge_label": edge_label,
        }

        #* for tkg and thg
        if ("edge_type" in df):
            edge_type = np.array(df["edge_type"]).astype(int)
            self._edge_type = edge_type
            full_data["edge_type"] = edge_type

        self._full_data = full_data

        if ("yago" in self.name):
            _train_mask, _val_mask, _test_mask = self.generate_splits(full_data, val_ratio=0.1, test_ratio=0.10) #99) #val_ratio=0.097, test_ratio=0.099)
        else:
            _train_mask, _val_mask, _test_mask = self.generate_splits(full_data, val_ratio=0.15, test_ratio=0.15)
        self._train_mask = _train_mask
        self._val_mask = _val_mask
        self._test_mask = _test_mask

    def generate_splits(
        self,
        full_data: Dict[str, Any],
        val_ratio: float = 0.15,
        test_ratio: float = 0.15,
    ) -> Tuple[Dict[str, Any], Dict[str, Any], Dict[str, Any]]:
        r"""Generates train, validation, and test splits from the full dataset
        Args:
            full_data: dictionary containing the full dataset
            val_ratio: ratio of validation data
            test_ratio: ratio of test data
        Returns:
            train_data: dictionary containing the training dataset
            val_data: dictionary containing the validation dataset
            test_data: dictionary containing the test dataset
        """
        val_time, test_time = list(
            np.quantile(
                full_data["timestamps"],
                [(1 - val_ratio - test_ratio), (1 - test_ratio)],
            )
        )
        timestamps = full_data["timestamps"]

        train_mask = timestamps <= val_time
        val_mask = np.logical_and(timestamps <= test_time, timestamps > val_time)
        test_mask = timestamps > test_time

        return train_mask, val_mask, test_mask
    
    def preprocess_static_edges(self):
        """
        Pre-process the static edges of the dataset
        """
        if ("staticfile" in self.meta_dict):
            OUT_DF = self.root + "/" + "ml_{}.pkl".format(self.name + "_static")
            if (osp.exists(OUT_DF)) and (self.version_passed is True):
                print("loading processed file")
                static_dict = load_pkl(OUT_DF)
                self._static_data = static_dict
            else:
                print("file not processed, generating processed file")
                static_dict, node_ids =  csv_to_staticdata(self.meta_dict["staticfile"], self._node_id)
                save_pkl(static_dict, OUT_DF)
                self._static_data = static_dict
        else:
            print ("static edges are only for tkgl-wikidata and tkgl-smallpedia datasets")

    
    @property
    def eval_metric(self) -> str:
        """
        the official evaluation metric for the dataset, loaded from info.py
        Returns:
            eval_metric: str, the evaluation metric
        """
        return self.metric

    @property
    def negative_sampler(self) -> NegativeEdgeSampler:
        r"""
        Returns the negative sampler of the dataset, will load negative samples from disc
        Returns:
            negative_sampler: NegativeEdgeSampler
        """
        return self.ns_sampler
    

    def load_val_ns(self) -> None:
        r"""
        load the negative samples for the validation set
        """
        self.ns_sampler.load_eval_set(
            fname=self.meta_dict["val_ns"], split_mode="val"
        )

    def load_test_ns(self) -> None:
        r"""
        load the negative samples for the test set
        """
        self.ns_sampler.load_eval_set(
            fname=self.meta_dict["test_ns"], split_mode="test"
        )

    @property
    def num_nodes(self) -> int:
        r"""
        Returns the total number of unique nodes in the dataset 
        Returns:
            num_nodes: int, the number of unique nodes
        """
        src = self._full_data["sources"]
        dst = self._full_data["destinations"]
        all_nodes = np.concatenate((src, dst), axis=0)
        uniq_nodes = np.unique(all_nodes, axis=0)
        return uniq_nodes.shape[0]
    

    @property
    def num_edges(self) -> int:
        r"""
        Returns the total number of edges in the dataset
        Returns:
            num_edges: int, the number of edges
        """
        src = self._full_data["sources"]
        return src.shape[0]
    

    @property
    def num_rels(self) -> int:
        r"""
        Returns the number of relation types in the dataset
        Returns:
            num_rels: int, the number of relation types
        """
        #* if it is a homogenous graph
        if ("edge_type" not in self._full_data):
            return 1
        else:
            return np.unique(self._full_data["edge_type"]).shape[0]

    @property
    def node_feat(self) -> Optional[np.ndarray]:
        r"""
        Returns the node features of the dataset with dim [N, feat_dim]
        Returns:
            node_feat: np.ndarray, [N, feat_dim] or None if there is no node feature
        """
        return self._node_feat
    
    @property
    def node_type(self) -> Optional[np.ndarray]:
        r"""
        Returns the node types of the dataset with dim [N], only for temporal heterogeneous graphs
        Returns:
            node_feat: np.ndarray, [N] or None if there is no node feature
        """
        return self._node_type

    @property
    def edge_feat(self) -> Optional[np.ndarray]:
        r"""
        Returns the edge features of the dataset with dim [E, feat_dim]
        Returns:
            edge_feat: np.ndarray, [E, feat_dim] or None if there is no edge feature
        """
        return self._edge_feat
    
    @property
    def edge_type(self) -> Optional[np.ndarray]:
        r"""
        Returns the edge types of the dataset with dim [E, 1], only for temporal knowledge graph and temporal heterogeneous graph
        Returns:
            edge_type: np.ndarray, [E, 1] or None if it is not a TKG or THG
        """
        return self._edge_type
    
    @property
    def static_data(self) -> Optional[np.ndarray]:
        r"""
        Returns the static edges related to this dataset, applies for tkgl-wikidata and tkgl-smallpedia, edges are (src, dst, rel_type)
        Returns:
            df: pd.DataFrame {"head": np.ndarray, "tail": np.ndarray, "rel_type": np.ndarray}
        """
        if (self.name == "tkgl-wikidata") or (self.name == "tkgl-smallpedia"):
            self.preprocess_static_edges()
        return self._static_data

    @property
    def full_data(self) -> Dict[str, Any]:
        r"""
        the full data of the dataset as a dictionary with keys: 'sources', 'destinations', 'timestamps', 'edge_idxs', 'edge_feat', 'w', 'edge_label',

        Returns:
            full_data: Dict[str, Any]
        """
        if self._full_data is None:
            raise ValueError(
                "dataset has not been processed yet, please call pre_process() first"
            )
        return self._full_data

    @property
    def train_mask(self) -> np.ndarray:
        r"""
        Returns the train mask of the dataset
        Returns:
            train_mask: training masks
        """
        if self._train_mask is None:
            raise ValueError("training split hasn't been loaded")
        return self._train_mask

    @property
    def val_mask(self) -> np.ndarray:
        r"""
        Returns the validation mask of the dataset
        Returns:
            val_mask: Dict[str, Any]
        """
        if self._val_mask is None:
            raise ValueError("validation split hasn't been loaded")
        return self._val_mask

    @property
    def test_mask(self) -> np.ndarray:
        r"""
        Returns the test mask of the dataset:
        Returns:
            test_mask: Dict[str, Any]
        """
        if self._test_mask is None:
            raise ValueError("test split hasn't been loaded")
        return self._test_mask


def main():

    name = "tkgl-polecat"
    dataset = LinkPropPredDataset(name=name, root="datasets", preprocess=True)
    dataset.edge_type



    # name = "tgbl-comment" 
    # dataset = LinkPropPredDataset(name=name, root="datasets", preprocess=True)

    # dataset.node_feat
    # dataset.edge_feat  # not the edge weights
    # dataset.full_data
    # dataset.full_data["edge_idxs"]
    # dataset.full_data["sources"]
    # dataset.full_data["destinations"]
    # dataset.full_data["timestamps"]
    # dataset.full_data["edge_label"]


if __name__ == "__main__":
    main()



"""
Sample and Generate negative edges that are going to be used for evaluation of a dynamic graph learning model
Negative samples are generated and saved to files ONLY once; 
    other times, they should be loaded from file with instances of the `negative_sampler.py`.
"""

import torch
import random
from torch import Tensor
import numpy as np
from torch_geometric.data import TemporalData
from tgb.utils.utils import save_pkl, load_pkl
from tgb.utils.info import PROJ_DIR
import os.path as osp
import os
import time
from tqdm import tqdm

from torch_geometric.datasets import JODIEDataset
from torch_geometric.loader import TemporalDataLoader

from tgb.linkproppred.dataset_pyg import PyGLinkPropPredDataset


class NegativeEdgeGenerator(object):
    def __init__(
        self,
        dataset_name: str,
        first_dst_id: int,
        last_dst_id: int,
        num_neg_e: int = 100,  # number of negative edges sampled per positive edges --> make it constant => 1000
        strategy: str = "rnd",
        rnd_seed: int = 123,
        hist_ratio: float = 0.5,
        historical_data: TemporalData = None,
    ) -> None:
        r"""
        Negative Edge Sampler class
        this is a class for generating negative samples for a specific datasets
        the set of the positive samples are provided, the negative samples are generated with specific strategies 
        and are saved for consistent evaluation across different methods
        negative edges are sampled with 'oen_vs_many' strategy.
        it is assumed that the destination nodes are indexed sequentially with 'first_dst_id' 
        and 'last_dst_id' being the first and last index, respectively.

        Parameters:
            dataset_name: name of the dataset
            first_dst_id: identity of the first destination node
            last_dst_id: indentity of the last destination node
            num_neg_e: number of negative edges being generated per each positive edge
            strategy: how to generate negative edges; can be 'rnd' or 'hist_rnd'
            rnd_seed: random seed for consistency
            hist_ratio: if the startegy is 'hist_rnd', how much of the negatives are historical
            historical_data: previous records of the positive edges
        
        Returns:
            None
        """
        self.rnd_seed = rnd_seed
        np.random.seed(self.rnd_seed)
        self.dataset_name = dataset_name

        self.first_dst_id = first_dst_id
        self.last_dst_id = last_dst_id
        self.num_neg_e = num_neg_e
        assert strategy in [
            "rnd",
            "hist_rnd",
        ], "The supported strategies are `rnd` or `hist_rnd`!"
        self.strategy = strategy
        if self.strategy == "hist_rnd":
            assert (
                historical_data != None
            ), "Train data should be passed when `hist_rnd` strategy is selected."
            self.hist_ratio = hist_ratio
            self.historical_data = historical_data

    def generate_negative_samples(self, 
                                  data: TemporalData, 
                                  split_mode: str, 
                                  partial_path: str,
                                  ) -> None:
        r"""
        Generate negative samples

        Parameters:
            data: an object containing positive edges information
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits
            partial_path: in which directory save the generated negatives
        """
        # file name for saving or loading...
        filename = (
            partial_path
            + "/"
            + self.dataset_name
            + "_"
            + split_mode
            + "_"
            + "ns"
            + ".pkl"
        )

        if self.strategy == "rnd":
            self.generate_negative_samples_rnd(data, split_mode, filename)
        elif self.strategy == "hist_rnd":
            self.generate_negative_samples_hist_rnd(
                self.historical_data, data, split_mode, filename
            )
        else:
            raise ValueError("Unsupported negative sample generation strategy!")

    def generate_negative_samples_rnd(self, 
                                      data: TemporalData, 
                                      split_mode: str, 
                                      filename: str,
                                      ) -> None:
        r"""
        Generate negative samples based on the `HIST-RND` strategy:
            - for each positive edge, sample a batch of negative edges from all possible edges with the same source node
            - filter actual positive edges
        
        Parameters:
            data: an object containing positive edges information
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits
            filename: name of the file containing the generated negative edges
        """
        print(
            f"INFO: Negative Sampling Strategy: {self.strategy}, Data Split: {split_mode}"
        )
        assert split_mode in [
            "val",
            "test",
        ], "Invalid split-mode! It should be `val` or `test`!"

        if os.path.exists(filename):
            print(
                f"INFO: negative samples for '{split_mode}' evaluation are already generated!"
            )
        else:
            print(f"INFO: Generating negative samples for '{split_mode}' evaluation!")
            # retrieve the information from the batch
            pos_src, pos_dst, pos_timestamp = (
                data.src.cpu().numpy(),
                data.dst.cpu().numpy(),
                data.t.cpu().numpy(),
            )

            # all possible destinations
            all_dst = np.arange(self.first_dst_id, self.last_dst_id + 1)

            evaluation_set = {}
            # generate a list of negative destinations for each positive edge
            pos_edge_tqdm = tqdm(
                zip(pos_src, pos_dst, pos_timestamp), total=len(pos_src)
            )
            for (
                pos_s,
                pos_d,
                pos_t,
            ) in pos_edge_tqdm:
                t_mask = pos_timestamp == pos_t
                src_mask = pos_src == pos_s
                fn_mask = np.logical_and(t_mask, src_mask)
                pos_e_dst_same_src = pos_dst[fn_mask]
                filtered_all_dst = np.setdiff1d(all_dst, pos_e_dst_same_src)

                '''
                when num_neg_e is larger than all possible destinations simple return all possible destinations
                '''
                if (self.num_neg_e > len(filtered_all_dst)):
                    neg_d_arr = filtered_all_dst
                else:
                    neg_d_arr = np.random.choice(
                    filtered_all_dst, self.num_neg_e, replace=False) #never replace negatives

                evaluation_set[(pos_s, pos_d, pos_t)] = neg_d_arr

            # save the generated evaluation set to disk
            save_pkl(evaluation_set, filename)

    def generate_historical_edge_set(self, 
                                     historical_data: TemporalData,
                                     ) -> tuple:
        r"""
        Generate the set of edges seen durign training or validation

        ONLY `train_data` should be passed as historical data; i.e., the HISTORICAL negative edges should be selected from training data only.
        
        Parameters:
            historical_data: contains the positive edges observed previously
        
        Returns:
            historical_edges: distict historical positive edges
            hist_edge_set_per_node: historical edges observed for each node
        """
        sources = historical_data.src.cpu().numpy()
        destinations = historical_data.dst.cpu().numpy()
        historical_edges = {}
        hist_e_per_node = {}
        for src, dst in zip(sources, destinations):
            # edge-centric
            if (src, dst) not in historical_edges:
                historical_edges[(src, dst)] = 1

            # node-centric
            if src not in hist_e_per_node:
                hist_e_per_node[src] = [dst]
            else:
                hist_e_per_node[src].append(dst)

        hist_edge_set_per_node = {}
        for src, dst_list in hist_e_per_node.items():
            hist_edge_set_per_node[src] = np.array(list(set(dst_list)))

        return historical_edges, hist_edge_set_per_node

    def generate_negative_samples_hist_rnd(
        self, 
        historical_data : TemporalData, 
        data: TemporalData, 
        split_mode: str, 
        filename: str,
    ) -> None:
        r"""
        Generate negative samples based on the `HIST-RND` strategy:
            - up to 50% of the negative samples are selected from the set of edges seen during the training with the same source node.
            - the rest of the negative edges are randomly sampled with the fixed source node.

        Parameters:
            historical_data: contains the history of the observed positive edges including 
                            distinct positive edges and edges observed for each positive node
            data: an object containing positive edges information
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits
            filename: name of the file to save generated negative edges
        
        Returns:
            None
        """
        print(
            f"INFO: Negative Sampling Strategy: {self.strategy}, Data Split: {split_mode}"
        )
        assert split_mode in [
            "val",
            "test",
        ], "Invalid split-mode! It should be `val` or `test`!"

        if os.path.exists(filename):
            print(
                f"INFO: negative samples for '{split_mode}' evaluation are already generated!"
            )
        else:
            print(f"INFO: Generating negative samples for '{split_mode}' evaluation!")
            # retrieve the information from the batch
            pos_src, pos_dst, pos_timestamp = (
                data.src.cpu().numpy(),
                data.dst.cpu().numpy(),
                data.t.cpu().numpy(),
            )

            pos_ts_edge_dict = {} #{ts: {src: [dsts]}}
            pos_edge_tqdm = tqdm(
                zip(pos_src, pos_dst, pos_timestamp), total=len(pos_src)
            )
            for (
                pos_s,
                pos_d,
                pos_t,
            ) in pos_edge_tqdm:
                if (pos_t not in pos_ts_edge_dict):
                    pos_ts_edge_dict[pos_t] = {pos_s: [pos_d]}
                else:
                    if (pos_s not in pos_ts_edge_dict[pos_t]):
                        pos_ts_edge_dict[pos_t][pos_s] = [pos_d]
                    else:
                        pos_ts_edge_dict[pos_t][pos_s].append(pos_d)

            # all possible destinations
            all_dst = np.arange(self.first_dst_id, self.last_dst_id + 1)

            # get seen edge history
            (
                historical_edges,
                hist_edge_set_per_node,
            ) = self.generate_historical_edge_set(historical_data)

            # sample historical edges
            max_num_hist_neg_e = int(self.num_neg_e * self.hist_ratio)

            evaluation_set = {}
            # generate a list of negative destinations for each positive edge
            pos_edge_tqdm = tqdm(
                zip(pos_src, pos_dst, pos_timestamp), total=len(pos_src)
            )
            for (
                pos_s,
                pos_d,
                pos_t,
            ) in pos_edge_tqdm:
                pos_e_dst_same_src = np.array(pos_ts_edge_dict[pos_t][pos_s])

                # sample historical edges
                num_hist_neg_e = 0
                neg_hist_dsts = np.array([])
                seen_dst = []
                if pos_s in hist_edge_set_per_node:
                    seen_dst = hist_edge_set_per_node[pos_s]
                    if len(seen_dst) >= 1:
                        filtered_all_seen_dst = np.setdiff1d(seen_dst, pos_e_dst_same_src)
                        #filtered_all_seen_dst = seen_dst #! no collision check
                        num_hist_neg_e = (
                            max_num_hist_neg_e
                            if max_num_hist_neg_e <= len(filtered_all_seen_dst)
                            else len(filtered_all_seen_dst)
                        )
                        neg_hist_dsts = np.random.choice(
                            filtered_all_seen_dst, num_hist_neg_e, replace=False
                        )

                # sample random edges
                if (len(seen_dst) >= 1):
                    invalid_dst = np.concatenate((np.array(pos_e_dst_same_src), seen_dst))
                else:
                    invalid_dst = np.array(pos_e_dst_same_src)
                filtered_all_rnd_dst = np.setdiff1d(all_dst, invalid_dst)

                num_rnd_neg_e = self.num_neg_e - num_hist_neg_e
                '''
                when num_neg_e is larger than all possible destinations simple return all possible destinations
                '''
                if (num_rnd_neg_e > len(filtered_all_rnd_dst)):
                    neg_rnd_dsts = filtered_all_rnd_dst
                else:
                    neg_rnd_dsts = np.random.choice(
                    filtered_all_rnd_dst, num_rnd_neg_e, replace=False
                )
                # concatenate the two sets: historical and random
                neg_dst_arr = np.concatenate((neg_hist_dsts, neg_rnd_dsts))
                evaluation_set[(pos_s, pos_d, pos_t)] = neg_dst_arr

            # save the generated evaluation set to disk
            save_pkl(evaluation_set, filename)



import dateutil.parser as dparser
import csv
import matplotlib.pyplot as plt
import numpy as np
from tqdm import tqdm
from os import listdir


def count_unique_countries(fname):
    node_dict = {}

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        # year,u,v,w
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                year = int(row[0])
                u = row[1]
                v = row[2]
                w = float(row[3])
                if u not in node_dict:
                    node_dict[u] = 1
                if v not in node_dict:
                    node_dict[v] = 1

    print("there are {} unique countries".format(len(node_dict)))



#! incorrect, do not use
def normalize_edgelist(fname: str, outname: str):
    """
    need to track id for nodes
    normalize the edgelist by row for each year
    """
    prev_t = 0
    uid = 0
    node_dict = {}
    year_dict = {}

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["year", "nation", "trading nation", "weight"]
        write.writerow(fields)
        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = int(row[0])
                    u = row[1]
                    v = row[2]
                    w = float(row[3])
                    if line_count == 1:
                        prev_t = ts
                    if u not in node_dict:
                        node_dict[u] = uid
                        uid += 1
                    if v not in node_dict:
                        node_dict[v] = uid
                        uid += 1
                    if w == 0:
                        line_count += 1
                        continue

                    if ts != prev_t:  # a new year now, write everything
                        # normalize the counts
                        for u in year_dict:
                            if np.sum(year_dict[u]) == 0:
                                continue
                            year_dict[u] = year_dict[u] / np.sum(year_dict[u])
                            invert_dict = {v: k for k, v in node_dict.items()}
                            for v in range(len(year_dict[u])):
                                if year_dict[u][v] > 0:
                                    write.writerow(
                                        [prev_t, u, invert_dict[v], year_dict[u][v]]
                                    )
                        year_dict = {}
                        prev_t = ts
                    else:
                        if u not in year_dict:
                            year_dict[u] = np.zeros(255)
                            year_dict[u][node_dict[v]] = w
                        else:
                            year_dict[u][node_dict[v]] = w
                    line_count += 1


def generate_aggregate_labels(fname: str, outname: str):
    """
    aggregate the node label for next year
    """

    ts_init = 1986

    # ts, src, subreddit, num_words, score
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["year", "nation", "trading nation", "weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = int(row[0])
                    u = row[1]
                    v = row[2]
                    w = float(row[3])
                    if (ts > ts_init):
                        write.writerow([ts, u, v, w])
                    line_count += 1


def check_sum_to_one(fname: str):
    """
    just to check if weights sum to 1 in a year
    """
    u_dict = {}
    ts_prev = 0

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        # ts, src, subreddit, num_words, score
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                ts = int(row[0])
                if (line_count == 1):
                    ts_prev = ts
                if (ts != ts_prev):
                    ts_prev = ts
                    for u in u_dict:
                        print (u_dict[u])
                    u_dict = {}
                u = row[1]
                v = row[2]
                w = float(row[3])
                if (u not in u_dict):
                    u_dict[u] = w
                else:
                    u_dict[u] += w
                line_count += 1




def main():
    #! should have the normalized version on the edgelist

    # #find the number of unique countries
    # fname = "un_trade_edgelist.csv"
    # count_unique_countries(fname)

    #! normalize edgelist by row for each year
    # fname = "un_trade_edgelist.csv"
    # outname = "un_trade_edgelist_normalized.csv"
    # normalize_edgelist(fname, outname)

    #! find the node label for next year
    # * the node labels are simply the edgelist in this case
    # fname = "un_trade_edgelist.csv"
    # outname = "un_trade_node_labels.csv"
    # generate_aggregate_labels(fname, outname)


    # #! check if all sums are correct
    # fname = "un_trade_node_labels.csv"
    # check_sum_to_one(fname)



if __name__ == "__main__":
    main()



"""
Sample and Generate negative edges that are going to be used for evaluation of a dynamic graph learning model
Negative samples are generated and saved to files ONLY once; 
    other times, they should be loaded from file with instances of the `negative_sampler.py`.
"""
import os
import torch
import numpy as np
from tqdm import tqdm
from torch_geometric.data import TemporalData
from tgb.utils.utils import save_pkl
from typing import Union


"""
negative sample generator for tkg datasets 
temporal filterted MRR
"""
class THGNegativeEdgeGenerator(object):
    def __init__(
        self,
        dataset_name: str,
        first_node_id: int,
        last_node_id: int,
        node_type: Union[np.ndarray, torch.Tensor],
        strategy: str = "node-type-filtered",
        num_neg_e: int = -1,  # -1 means generate all possible negatives
        rnd_seed: int = 1,
        edge_data: TemporalData = None,
    ) -> None:
        r"""
        Negative Edge Generator class for Temporal Heterogeneous Graphs
        this is a class for generating negative samples for a specific datasets
        the set of the positive samples are provided, the negative samples are generated with specific strategies 
        and are saved for consistent evaluation across different methods

        Parameters:
            dataset_name: name of the dataset
            first_node_id: the first node id
            last_node_id: the last node id
            node_type: the node type of each node
            strategy: the strategy to generate negative samples
            num_neg_e: number of negative samples to generate
            rnd_seed: random seed
            edge_data: the edge data object containing the positive edges
        Returns:
            None
        """
        self.rnd_seed = rnd_seed
        np.random.seed(self.rnd_seed)
        self.dataset_name = dataset_name
        self.first_node_id = first_node_id
        self.last_node_id = last_node_id
        if isinstance(node_type, torch.Tensor):
            node_type = node_type.cpu().numpy()
        self.node_type = node_type
        self.node_type_dict = self.get_destinations_based_on_node_type(first_node_id, last_node_id, self.node_type) # {node_type: {nid:1}}
        assert isinstance(self.node_type, np.ndarray), "node_type should be a numpy array"
        self.num_neg_e = num_neg_e  #-1 means generate all 

        assert strategy in [
            "node-type-filtered",
            "random",
        ], "The supported strategies are `node-type-filtered`"
        self.strategy = strategy
        self.edge_data = edge_data

    def get_destinations_based_on_node_type(self, 
                                            first_node_id: int,
                                            last_node_id: int,
                                            node_type: np.ndarray) -> dict:
        r"""
        get the destination node id arrays based on the node type
        Parameters:
            first_node_id: the first node id
            last_node_id: the last node id
            node_type: the node type of each node

        Returns:
            node_type_dict: a dictionary containing the destination node ids for each node type
        """
        node_type_store = {}
        assert first_node_id <= last_node_id, "Invalid destination node ids!"
        assert len(node_type) == (last_node_id - first_node_id + 1), "node type array must match the indices"
        for k in range(len(node_type)):
            nt = int(node_type[k]) #node type must be ints
            nid = k + first_node_id
            if nt not in node_type_store:
                node_type_store[nt] = {nid:1}
            else:
                node_type_store[nt][nid] = 1
        node_type_dict = {}
        for ntype in node_type_store:
            node_type_dict[ntype] = np.array(list(node_type_store[ntype].keys()))
            assert np.all(np.diff(node_type_dict[ntype]) >= 0), "Destination node ids for a given type must be sorted"
            assert np.all(node_type_dict[ntype] <= last_node_id), "Destination node ids must be less than or equal to the last destination id"
        return node_type_dict

    def generate_negative_samples(self, 
                                  pos_edges: TemporalData,
                                  split_mode: str, 
                                  partial_path: str,
                                  ) -> None:
        r"""
        Generate negative samples

        Parameters:
            pos_edges: positive edges to generate the negatives for
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits
            partial_path: in which directory save the generated negatives
        """
        # file name for saving or loading...
        filename = (
            partial_path
            + "/"
            + self.dataset_name
            + "_"
            + split_mode
            + "_"
            + "ns"
            + ".pkl"
        )

        if self.strategy == "node-type-filtered":
            self.generate_negative_samples_nt(pos_edges, split_mode, filename)
        elif self.strategy == "random":
            self.generate_negative_samples_random(pos_edges, split_mode, filename)
        else:
            raise ValueError("Unsupported negative sample generation strategy!")

    def generate_negative_samples_nt(self, 
                                      data: TemporalData, 
                                      split_mode: str, 
                                      filename: str,
                                      ) -> None:
        r"""
        now we consider (s, d, t, edge_type) as a unique edge, also adding the node type info for the destination node for convenience so (s, d, t, edge_type): (conflict_set, d_node_type)
        Generate negative samples based on the random strategy:
            - for each positive edge, retrieve all possible destinations based on the node type of the destination node
            - filter actual positive edges at the same timestamp with the same edge type
        
        Parameters:
            data: an object containing positive edges information
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits
            filename: name of the file containing the generated negative edges
        """
        print(
            f"INFO: Negative Sampling Strategy: {self.strategy}, Data Split: {split_mode}"
        )
        assert split_mode in [
            "val",
            "test",
        ], "Invalid split-mode! It should be `val` or `test`!"

        if os.path.exists(filename):
            print(
                f"INFO: negative samples for '{split_mode}' evaluation are already generated!"
            )
        else:
            print(f"INFO: Generating negative samples for '{split_mode}' evaluation!")
            # retrieve the information from the batch
            pos_src, pos_dst, pos_timestamp, edge_type = (
                data.src.cpu().numpy(),
                data.dst.cpu().numpy(),
                data.t.cpu().numpy(),
                data.edge_type.cpu().numpy(),
            )

            # generate a list of negative destinations for each positive edge
            pos_edge_tqdm = tqdm(
                zip(pos_src, pos_dst, pos_timestamp, edge_type), total=len(pos_src)
            )

            edge_t_dict = {} # {(t, u, edge_type): {v_1, v_2, ..} }
            #! iterate once to put all edges into a dictionary for reference
            for (
                pos_s,
                pos_d,
                pos_t,
                edge_type,
            ) in pos_edge_tqdm:
                if (pos_t, pos_s, edge_type) not in edge_t_dict:
                    edge_t_dict[(pos_t, pos_s, edge_type)] = {pos_d:1}
                else:
                    edge_t_dict[(pos_t, pos_s, edge_type)][pos_d] = 1

            out_dict = {}
            for key in tqdm(edge_t_dict):
                conflict_set = np.array(list(edge_t_dict[key].keys()))
                pos_d = conflict_set[0]
                #* retieve the node type of the destination node as well 
                #! assumption, same edge type = same destination node type
                d_node_type = int(self.node_type[pos_d - self.first_node_id])
                all_dst = self.node_type_dict[d_node_type]
                if (self.num_neg_e == -1):
                    filtered_all_dst = np.setdiff1d(all_dst, conflict_set)
                else:
                    #* lazy sampling
                    neg_d_arr = np.random.choice(
                        all_dst, self.num_neg_e, replace=False) #never replace negatives
                    if len(np.setdiff1d(neg_d_arr, conflict_set)) < self.num_neg_e:
                        neg_d_arr = np.random.choice(
                            np.setdiff1d(all_dst, conflict_set), self.num_neg_e, replace=False)
                    filtered_all_dst = neg_d_arr
                out_dict[key] = filtered_all_dst
            print ("ns samples for ", len(out_dict), " positive edges are generated")
            # save the generated evaluation set to disk
            save_pkl(out_dict, filename)

    def generate_negative_samples_random(self, 
                                      data: TemporalData, 
                                      split_mode: str, 
                                      filename: str,
                                      ) -> None:
        r"""
        generate random negative edges for ablation study
        
        Parameters:
            data: an object containing positive edges information
            split_mode: specifies whether to generate negative edges for 'validation' or 'test' splits
            filename: name of the file containing the generated negative edges
        """
        print(
            f"INFO: Negative Sampling Strategy: {self.strategy}, Data Split: {split_mode}"
        )
        assert split_mode in [
            "val",
            "test",
        ], "Invalid split-mode! It should be `val` or `test`!"

        if os.path.exists(filename):
            print(
                f"INFO: negative samples for '{split_mode}' evaluation are already generated!"
            )
        else:
            print(f"INFO: Generating negative samples for '{split_mode}' evaluation!")
            # retrieve the information from the batch
            pos_src, pos_dst, pos_timestamp, edge_type = (
                data.src.cpu().numpy(),
                data.dst.cpu().numpy(),
                data.t.cpu().numpy(),
                data.edge_type.cpu().numpy(),
            )
            first_dst_id = self.edge_data.dst.min()
            last_dst_id = self.edge_data.dst.max()
            all_dst = np.arange(first_dst_id, last_dst_id + 1)
            evaluation_set = {}
            # generate a list of negative destinations for each positive edge
            pos_edge_tqdm = tqdm(
                zip(pos_src, pos_dst, pos_timestamp, edge_type), total=len(pos_src)
            )

            for (
                pos_s,
                pos_d,
                pos_t,
                edge_type,
            ) in pos_edge_tqdm:
                t_mask = pos_timestamp == pos_t
                src_mask = pos_src == pos_s
                fn_mask = np.logical_and(t_mask, src_mask)
                pos_e_dst_same_src = pos_dst[fn_mask]
                filtered_all_dst = np.setdiff1d(all_dst, pos_e_dst_same_src)
                if (self.num_neg_e > len(filtered_all_dst)):
                    neg_d_arr = filtered_all_dst
                else:
                    neg_d_arr = np.random.choice(
                    filtered_all_dst, self.num_neg_e, replace=False) #never replace negatives
                evaluation_set[(pos_t, pos_s, edge_type)] = neg_d_arr
            save_pkl(evaluation_set, filename)












import os.path as osp
import os

r"""
General space to store global information used elsewhere such as url links, evaluation metrics etc.
"""
PROJ_DIR = osp.dirname(osp.abspath(os.path.join(__file__, os.pardir))) + "/"


class BColors:
    """
    A class to change the colors of the strings.
    """

    HEADER = "\033[95m"
    OKBLUE = "\033[94m"
    OKCYAN = "\033[96m"
    OKGREEN = "\033[92m"
    WARNING = "\033[93m"
    FAIL = "\033[91m"
    ENDC = "\033[0m"
    BOLD = "\033[1m"
    UNDERLINE = "\033[4m"

DATA_URL_DICT = {
    "tgbl-enron": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-enron.zip",
    "tgbl-uci": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-uci.zip",
    "tgbl-wiki":"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-wiki-v2.zip", #"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-wiki.zip", #v1
    "tgbl-subreddit":"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-subreddit.zip",
    "tgbl-lastfm":"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-lastfm.zip",
    "tgbl-review": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-review-v2.zip",  # "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-review-v3.zip" #"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-review-v2.zip"  #"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-review.zip", #v1
    "tgbl-coin": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-coin-v2.zip", #"https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-coin.zip",
    "tgbl-flight": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-flight-v2.zip", #"tgbl-flight": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-flight_edgelist_v2_ts.zip",
    "tgbl-comment": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbl-comment.zip",
    "tgbn-trade": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbn-trade.zip",
    "tgbn-genre": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbn-genre.zip",
    "tgbn-reddit": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbn-reddit.zip",
    "tgbn-token": "https://object-arbutus.cloud.computecanada.ca/tgb/tgbn-token.zip",
    "tkgl-polecat": "https://object-arbutus.cloud.computecanada.ca/tgb/tkgl-polecat.zip",
    "tkgl-icews": "https://object-arbutus.cloud.computecanada.ca/tgb/tkgl-icews.zip",
    "tkgl-yago":"https://object-arbutus.cloud.computecanada.ca/tgb/tkgl-yago.zip",
    "tkgl-wikidata": "https://object-arbutus.cloud.computecanada.ca/tgb/tkgl-wikidata.zip",
    "tkgl-smallpedia": "https://object-arbutus.cloud.computecanada.ca/tgb/tkgl-smallpedia.zip",
    "thgl-myket": "https://object-arbutus.cloud.computecanada.ca/tgb/thgl-myket.zip",
    "thgl-github": "https://object-arbutus.cloud.computecanada.ca/tgb/thgl-github.zip",
    "thgl-forum": "https://object-arbutus.cloud.computecanada.ca/tgb/thgl-forum.zip",
    "thgl-software": "https://object-arbutus.cloud.computecanada.ca/tgb/thgl-software.zip", #"https://object-arbutus.cloud.computecanada.ca/tgb/thgl-software_ns_random.zip"
}



DATA_VERSION_DICT = {
    "tgbl-enron": 1,
    "tgbl-uci": 1,
    "tgbl-wiki": 2,  
    "tgbl-subreddit": 1,
    "tgbl-lastfm": 1,
    "tgbl-review": 2, #3
    "tgbl-coin": 2,
    "tgbl-comment": 1,
    "tgbl-flight": 2,
    "tgbn-trade": 1,
    "tgbn-genre": 1,
    "tgbn-reddit": 1,
    "tgbn-token": 1,
    "tkgl-polecat": 1,
    "tkgl-icews": 1,
    "tkgl-yago": 1,
    "tkgl-wikidata": 1,
    "tkgl-smallpedia": 1,
    "thgl-myket": 1,
    "thgl-github": 1,
    "thgl-forum": 1,
    "thgl-software": 1,
}


DATA_EVAL_METRIC_DICT = {
    "tgbl-enron": "mrr",
    "tgbl-uci": "mrr",
    "tgbl-wiki": "mrr",
    "tgbl-subreddit": "mrr",
    "tgbl-lastfm": "mrr",
    "tgbl-review": "mrr",
    "tgbl-coin": "mrr",
    "tgbl-comment": "mrr",
    "tgbl-flight": "mrr",
    "tkgl-polecat": "mrr",
    "tkgl-yago": "mrr",
    "tkgl-wikidata": "mrr",
    "tkgl-smallpedia": "mrr",
    "tkgl-icews": "mrr",
    "thgl-myket": "mrr",
    "thgl-github": "mrr",
    "thgl-forum": "mrr",
    "thgl-software": "mrr",
    "tgbn-trade": "ndcg",
    "tgbn-genre": "ndcg",
    "tgbn-reddit": "ndcg",
    "tgbn-token": "ndcg",
}

DATA_NS_STRATEGY_DICT = {
    "tgbl-enron": "hist_rnd",
    "tgbl-uci": "hist_rnd",
    "tgbl-wiki": "hist_rnd",
    "tgbl-subreddit": "hist_rnd",
    "tgbl-lastfm": "hist_rnd",
    "tgbl-review": "hist_rnd",
    "tgbl-coin": "hist_rnd",
    "tgbl-comment": "hist_rnd",
    "tgbl-flight": "hist_rnd",
    "tkgl-polecat": "time-filtered",
    "tkgl-yago": "time-filtered",
    "tkgl-wikidata": "dst-time-filtered",
    "tkgl-smallpedia": "time-filtered",
    "tkgl-icews": "time-filtered",
    "thgl-myket": "node-type-filtered",
    "thgl-github": "node-type-filtered",
    "thgl-forum": "node-type-filtered",
    "thgl-software": "node-type-filtered",
}


DATA_NUM_CLASSES = {
    "tgbn-trade": 255,
    "tgbn-genre": 513,
    "tgbn-reddit": 698,
    "tgbn-token": 1001,
}



import csv


with open('ml_enron.csv', 'r', newline='\n') as infile, open('tgbl-enron_edgelist.csv', 'w', newline='\n') as outfile:
    reader = csv.reader(infile)
    writer = csv.writer(outfile)
    for row in reader:
        writer.writerow(row[1:])


import csv
from tqdm import tqdm
from os import listdir
from tgb.utils.stats import analyze_csv


def find_filenames(path_to_dir):
    r"""
    find all files in a folder
    Parameters:
        path_to_dir (str): path to the directory
    """
    filenames = listdir(path_to_dir)
    return filenames


def read_edgelist(fname, outfname, write_header=False):
    """
    read a space separated edgelist
    comment’s author, author of the parent (the post that the comment is replied to), comment’s creation time, comment’s edge id
    u,v,t,edge_id
    3746738	1637382	1551398391	31534079835
    Parameters:
        fname (str): path to the edgelist
        outfname (str): path to the output file
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    with open(outfname, "a") as outf:
        write = csv.writer(outf)
        if write_header:
            fields = ["ts", "src", "dst", "edge_id"]
            write.writerow(fields)
        for line in lines:
            line = line.split()
            if len(line) < 4:
                continue
            src = line[0]
            dst = line[1]
            ts = line[2]
            edge_id = line[3]
            write.writerow([ts, src, dst, edge_id])


def read_nodeattr(fname, outfname, write_header=False):
    """
    read a space separated edgelist
    comment’s edge id, Reddit’s identifier of the comment, Reddit’s identifier of the parent (the post that the comment is replied to)
    Reddit’s identifier of the submission that the comment is in, name of the subreddit that the comment is in, number of characters in the comment’s body
    number of words in the comment’s body, score of the comment, a flag indicating if the comment has been edited


    edge_id, subreddit, num_characters, num_words, score, 'edited_flag'
    Parameters:
        fname (str): path to the edgelist
        outfname (str): path to the output file
    """
    edgelist = open(fname, "r")
    lines = list(edgelist.readlines())
    edgelist.close()

    with open(outfname, "a") as outf:
        write = csv.writer(outf)
        if write_header:
            fields = [
                "edge_id",
                "subreddit",
                "num_characters",
                "num_words",
                "score",
                "edited_flag",
            ]
            write.writerow(fields)
        for line in lines:
            line = line.split()
            if len(line) < 4:
                continue
            edge_id = line[0]
            subreddit = line[4]
            num_characters = line[5]
            num_words = line[6]
            score = line[7]
            edited_flag = line[8].strip("/n")
            write.writerow(
                [edge_id, subreddit, num_characters, num_words, score, edited_flag]
            )


def combine_edgelist_edgefeat(edgefname, featfname, outname):
    """
    combine edgelist and edge features
    #! remove subreddit from feature
    """
    total_lines = sum(1 for line in open(edgefname))
    subreddit_ids = {}

    missing_ts = 0
    missing_src = 0
    missing_dst = 0
    line_idx = 0

    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "src", "dst", "subreddit", "num_words", "score"]
        write.writerow(fields)
        sub_id = 0
        edgelist = open(edgefname, "r")
        edgefeat = open(featfname, "r")
        edgelist.readline()
        edgefeat.readline()

        while True:
            #'ts', 'src', 'dst', 'edge_id'
            edge_line = edgelist.readline()
            edge_line = edge_line.split(",")
            if len(edge_line) < 4:
                break
            edge_id = int(edge_line[3])
            ts = int(edge_line[0])
            src = int(edge_line[1])
            dst = int(edge_line[2])

            #'edge_id', 'subreddit', 'num_characters', 'num_words', 'score', 'edited_flag'
            feat_line = edgefeat.readline()
            feat_line = feat_line.split(",")
            edge_id_feat = int(feat_line[0])
            subreddit = feat_line[1]
            if subreddit not in subreddit_ids:
                subreddit_ids[subreddit] = sub_id
                sub_id += 1
            subreddit = subreddit_ids[subreddit]
            num_characters = int(feat_line[2])
            num_words = int(feat_line[3])
            score = int(feat_line[4])
            edited_flag = bool(feat_line[5])

            #! check if ts, src, dst is -1
            if ts == -1:
                missing_ts += 1
                continue
            if src == -1:
                missing_src += 1
                continue
            if dst == -1:
                missing_dst += 1
                continue

            if edge_id != edge_id_feat:
                print("edge_id != edge_id_feat")
                print(edge_id)
                print(edge_id_feat)
                break

            # write.writerow([ts, src, dst, subreddit, num_words, score])
            write.writerow([ts, src, dst, num_words, score])
            line_idx += 1
    print("processed", line_idx, "lines")
    # print ("there are lines", missing_ts, " missing timestamps")
    # print ("there are lines", missing_src, " missing src")
    # print ("there are lines", missing_dst, " missing dst")


def main():
    # #! unzip all xz files by $ unxz *.xz

    # f_dir = "raw/raw_2008_2010/" #"raw/raw_2005_2010/" #"raw/raw_2013_2014/"
    # fnames = find_filenames(f_dir)
    # outname = "redditcomments_edgelist_2008_2010.csv" #"redditcomments_edgelist_2013_2014.csv"
    # idx = 0
    # for fname in tqdm(fnames):
    #     if (idx == 0):
    #         read_edgelist(f_dir+fname, outname, write_header=True)
    #     else:
    #         read_edgelist(f_dir+fname, outname, write_header=False)
    #     idx += 1

    # # #! extract the node attributes
    f_dir = "raw/node_2008_2010/"#"raw/node_2005_2010/"
    fnames = find_filenames(f_dir)
    outname = "redditcomments_edgefeat_2008_2010.csv"
    idx = 0
    for fname in tqdm(fnames):
        if (idx == 0):
            read_nodeattr(f_dir+fname, outname, write_header=True)
        else:
            read_nodeattr(f_dir+fname, outname, write_header=False)
        idx += 1

    #! combine edgelist and edge feat file check if the edge_id matches
    # edgefname = "redditcomments_edgelist_2005_2010.csv"
    # featfname = "redditcomments_edgefeat_2005_2010.csv"
    # outname = "redditcomments_edgelist.csv"
    # combine_edgelist_edgefeat(edgefname, featfname, outname)

    # #! analyze the extracted csv
    # fname = "redditcomments_edgelist_2005_2010.csv"
    # analyze_csv(fname)


if __name__ == "__main__":
    main()



r"""
How to use
python tkgl_wikidata.py --chunk 0 --num_chunks 25
# python tkgl_wikidata.py --chunk 1 --num_chunks 25
"""

from qwikidata.entity import WikidataItem
from qwikidata.json_dump import WikidataJsonDump
from qwikidata.datavalue import get_datavalue_from_snak_dict, WikibaseEntityId
from tqdm import tqdm
from collections import defaultdict
import os.path as osp
import os
import pickle
import argparse
import numpy as np
import csv

def timeEdgeWrite2csv(outname, out_dict):
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['timestamp', 'head', 'tail', 'relation_type', 'time_rel_type'])
        for edge in out_dict.keys():
            ts = edge[0]
            src = edge[1]
            dst = edge[2]
            rel_type = edge[3]
            time_rel_type = edge[4]
            row = [ts, src, dst, rel_type, time_rel_type]
            writer.writerow(row)


def EdgeWrite2csv(outname, out_dict):
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['head', 'tail', 'relation_type'])
        for edge in out_dict.keys():
            src = edge[0]
            dst = edge[1]
            rel_type = edge[2]
            row = [src, dst, rel_type]
            writer.writerow(row)


def main():
    parser = argparse.ArgumentParser(description='Process some integers.')
    # parser.add_argument('--split', type=str, default = 'train',
    #                     help='an integer for the accumulator')
    parser.add_argument('--chunk', type=int, default = 0,
                        help='an integer for the accumulator')
    parser.add_argument('--num_chunks', type=int, default = 10,
                        help='an integer for the accumulator')

    args = parser.parse_args()
    print(args)
    assert args.chunk < args.num_chunks

    # # create an instance of WikidataJsonDump
    # if args.split == 'train':
    #     wjd_dump_path_original = "wikidata-20210517-all.json.gz"
    # elif args.split == 'val':
    #     wjd_dump_path_original = "wikidata-20210607-all.json.gz"
    # elif args.split == 'test':
    #     wjd_dump_path_original = 'wikidata-20210628-all.json.gz'
    # else:
    #     raise ValueError('Unknown split')

    #* download here
    #? https://dumps.wikimedia.org/wikidatawiki/entities/

    wjd_dump_path_original = "wikidata-20240220-all.json.gz" #"latest-all_03_Apr_2024_12_49.json" 

    wjd_dump_path = osp.join('dump', wjd_dump_path_original)
    wjd = WikidataJsonDump(wjd_dump_path)

    print(wjd_dump_path)

    

    """
    # head = entity_dict['id']
    # type = entity_dict['type']
    # labels = entity_dict['labels']
    # descriptions = entity_dict['descriptions']
    # aliases = entity_dict['aliases']
    # if ('claims' in entity_dict):
    #     claims = entity_dict['claims']
    #     for key in claims.keys():
    #         print (key)
    #         print (claims[key])
    # sitelinks = entity_dict['sitelinks']
    """

    time_edge_dict = {} #{()}
    time_rel_dict = {}
    static_edge_dict = {}
    dummy_rel_set = ['P31','P279']  #filter out instance of and subclass of
    time_rel_set = ['P585','P580', 'P582', 'P577', 'P574']  #point in time, start time, end time, publication date,year of publication of scientific name for taxon

    num_totals = 100000000 #4000000 #10000000 #110000000

    tmp = np.linspace(0, num_totals, args.num_chunks + 1).astype(np.int64)
    start_idx = tmp[args.chunk]
    end_idx = tmp[args.chunk + 1]
    print('Start: ', start_idx)
    print('End: ', end_idx)


    #? output format is (timestamp, head, tail, relation_type, time_rel_type)
    for i, entity_dict in enumerate(tqdm(wjd, total=(end_idx))):
        #! entity_dict keys(['type', 'id', 'labels', 'descriptions', 'aliases', 'claims', 'sitelinks', 'pageid', 'ns', 'title', 'lastrevid', 'modified'])
        if i > end_idx:
            break

        if not (start_idx <= i and i < end_idx):
            continue

        head = entity_dict['id']

        # head needs to start from 'Q'
        if head[0] == 'Q':
            head_id = head
            if 'claims' in entity_dict:
                claim_dict = entity_dict['claims']
                rel_list = list(claim_dict.keys())
                for rel in rel_list:
                    if (rel in dummy_rel_set):
                        continue
                    tail_list = claim_dict[rel]
                    for tail in tail_list:
                        tail_id = None
                        #* first check if there is a valid tail
                        if (tail['mainsnak']['datatype'] == 'wikibase-item'):
                            if ('rank' in tail) and (tail['rank'] != 'deprecated') and ('datavalue' in tail['mainsnak']):
                                if 'id' in tail['mainsnak']['datavalue']['value']:
                                    tail_id = tail['mainsnak']['datavalue']['value']['id']
                                else: 
                                    tail_id = 'Q' + str(tail['mainsnak']['datavalue']['value']['numeric-id'])

                        #* check if there is a qualifier and if it is a time qualifier
                        if (tail_id is not None):
                            if ("qualifiers" in tail):
                                time_logged = False
                                for q in tail["qualifiers"]:
                                    for item in tail["qualifiers"][q]:
                                        if (item['datatype'] == 'time') and ('datavalue' in item):
                                            timestr = item['datavalue']['value']['time']
                                            time_rel_type = q
                                            if (time_rel_type in time_rel_set):
                                                time_edge_dict[(timestr, head_id, tail_id, rel, time_rel_type)] = 1
                                                time_logged = True
                                            else:
                                                time_logged = False
                                if not time_logged:
                                    static_edge_dict[(head_id, tail_id, rel)] = 1
                            else:
                                static_edge_dict[(head_id, tail_id, rel)] = 1

    #! write edges to file
    print ("there are ", len(time_edge_dict), " temporal edges in the dataset")
    outname = "time_edgelist_" + str(args.chunk) + ".csv" #"tkgl-wikidata_time_edgelist.csv"
    timeEdgeWrite2csv(outname, time_edge_dict)


    print ("there are ", len(static_edge_dict), " static edges in the dataset")
    outname = "static_edgelist_" + str(args.chunk) + ".csv" #"tkgl-wikidata_static_edgelist.csv"
    EdgeWrite2csv(outname, static_edge_dict)
    

                
                    
if __name__ == '__main__':

    main()



"""
Data source: 
# https://surfdrive.surf.nl/files/index.php/s/M09RDerAMZrQy8q#editor

# https://dl.acm.org/doi/abs/10.1145/3487553.3524699

also see here: https://arxiv.org/pdf/1803.03697

#  Temporal Social Network Dataset of Reddit
### Dataset to accompany “A large-scale temporal analysis of user lifespan durability on the Reddit social media platform” (WWW 2022).

## Overview

This dataset consists of more than 6.7 billion Reddit comment interactions made from the beginning of Reddit in 2005 until the end of 2019. 

### Nodes

Nodes in the network represent users who posted at least one comment or one submission until the end of 2019 and have not deleted their accounts by the time of the data ingestion. 

Each user is assigned a unique identifier starting from 0, and -1 is the identifier of the node representing deleted users. The `nodes` file maintains the node-identifier assignment to each user’s username.

### Edges

For each month of our data, we maintain two separate files, an edge file that consists of temporal edges data and an attribute file that consists of attributes of each interaction. All these files are in a tab-separated format. The compressed edge files and compressed attribute files are available in the `edges` and the `attributes` directory, respectively. The name of files indicates the timeframe they belong to.

Each line in an edge file corresponds to a comment and includes:

 - comment’s author
 - author of the parent (the post that the comment is replied to)
 - comment’s creation time
 - comment’s edge id

Each line in an attribute file corresponds to the line with the same line number in the corresponding edge file and includes:

 - comment’s edge id
 - Reddit’s identifier of the comment
 - Reddit’s identifier of the parent (the post that the comment is replied to)
 - Reddit’s identifier of the submission that the comment is in
 - name of the subreddit that the comment is in
 - number of characters in the comment’s body
 - number of words in the comment’s body
 - score of the comment
 - a flag indicating if the comment has been edited

### Stats

Size (compressed): 125GB
Size (uncompressed): 652GB
Number of nodes: 62,402,844
Number of edges: 6,728,759,080

### Notes

Reddit banned the subreddit `/r/Incels` in November of 2017, and its data is no longer available via the Reddit API. This has resulted in the loss of score data for 119,111 comments made in October and November of 2017 in this subreddit. The affected entries have a null value as their score. 

## Citation

If you want to reuse this dataset, you can reference it as follows:

A. Nadiri and F.W. Takes, A large-scale temporal analysis of user lifespan durability on the Reddit social media platform, in Proceedings of the 28th ACM International Web Conference (TheWebConf) Workshops, 2022.

## Online repository

The dataset is available for download at [**LINK**](https://surfdrive.surf.nl/files/index.php/s/M09RDerAMZrQy8q)

## Acknowledgments

The dataset is constructed using data provided by [The Pushshift Reddit Dataset](https://ojs.aaai.org/index.php/ICWSM/article/view/7347)


"""



"""
ideas for temporal heterogenous graph in reddit data:

node types:
1. user
2. subreddit

edge types
1. user post in subreddit (top level)
2. user replies to another user 
3. user replies in subreddit



# node types:
# 1. user 
# 2. subreddit
# 3. comment


# edge types
# 1. user makes comment in subreddit (top level comment)
# 2. user replies to comment in subreddit (comments that has a parent)
# 2. comment is child of comment (comments that has a parent)
# 3. comment belongs to subreddit
"""
import csv
from tgb.utils.utils import save_pkl, load_pkl


def load_csv_raw(fname):
    """
    load the raw csv file and merge them into one
    ts, src, dst, subreddit, reddit_id, reddit_parent_id, num_words, score
    """
    out_dict = {}
    num_lines = 0
    max_words = 0
    min_words = 10000000

    max_score = 0
    min_score = 1000000

    """
    relation types:
    0: user replies to user
    1: user replies to subreddit

    node types:
    0: user
    1: subreddit
    """

    node_dict = {}
    node_type_dict = {}
    reddit_deg_dict = {}
    node_deg_dict = {}
    header = True
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        #* ts, src, dst, subreddit, reddit_id, reddit_parent_id, num_words, score
        #? 1388534400,32183137,51851117,AskReddit,t1_ceefsvy,t3_1u4kbf,32,1
        for row in reader: 
            if header:
                header = False
                continue
            ts = int(row[0])
            src = row[1]
            if (src not in node_dict):
                node_dict[src] = len(node_dict)
                node_type_dict[node_dict[src]] = 0
            dst = row[2]
            if (dst not in node_dict):
                node_dict[dst] = len(node_dict)
                node_type_dict[node_dict[dst]] = 0

            if (src not in node_deg_dict):
                node_deg_dict[src] = 1
            else:
                node_deg_dict[src] += 1
            if (dst not in node_deg_dict):
                node_deg_dict[dst] = 1
            else:
                node_deg_dict[dst] += 1

            subreddit = row[3]
            if (subreddit not in node_dict):
                node_dict[subreddit] = len(node_dict)
                node_type_dict[node_dict[subreddit]] = 1
            
            if (subreddit not in reddit_deg_dict):
                reddit_deg_dict[subreddit] = 1
            else:
                reddit_deg_dict[subreddit] += 1
            
            num_words = int(row[6])
            if (num_words > max_words):
                max_words = num_words
            if (num_words < min_words):
                min_words = num_words
            score = int(row[7])
            if (score > max_score):
                max_score = score
            if (score < min_score):
                min_score = score

            if (ts in out_dict):
                out_dict[ts][(node_dict[src], node_dict[dst], 0)] = (num_words, score)
                out_dict[ts][(node_dict[src], node_dict[subreddit], 1)] = (num_words, score)
            else:
                out_dict[ts] = {}
                out_dict[ts][(node_dict[src], node_dict[dst], 0)] = (num_words, score)
                out_dict[ts][(node_dict[src], node_dict[subreddit], 1)] = (num_words, score)
            num_lines += 1

    print ("max words: ", max_words)
    print ("min words: ", min_words)
    print ("max score: ", max_score)
    print ("min score: ", min_score)
    return out_dict, num_lines, node_dict, node_type_dict, reddit_deg_dict, node_deg_dict



def load_csv_filtered_node(fname, low_deg_dict):
    """
    load the raw csv file, remove edges with low degree nodes
    ts, src, dst, subreddit, reddit_id, reddit_parent_id, num_words, score
    """
    out_dict = {}
    num_lines = 0
    """
    relation types:
    0: user replies to user
    1: user replies to subreddit

    node types:
    0: user
    1: subreddit
    """

    node_dict = {}
    node_type_dict = {}
    header = True
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter =',')
        #* ts, src, dst, subreddit, reddit_id, reddit_parent_id, num_words, score
        #? 1388534400,32183137,51851117,AskReddit,t1_ceefsvy,t3_1u4kbf,32,1
        for row in reader: 
            if header:
                header = False
                continue
            ts = int(row[0])
            src = row[1]
            dst = row[2]

            #* filter low degree nodes
            if (src in low_deg_dict or dst in low_deg_dict):
                continue

            if (src not in node_dict):
                node_dict[src] = len(node_dict)
                node_type_dict[node_dict[src]] = 0
            if (dst not in node_dict):
                node_dict[dst] = len(node_dict)
                node_type_dict[node_dict[dst]] = 0

            subreddit = row[3]
            if (subreddit not in node_dict):
                node_dict[subreddit] = len(node_dict)
                node_type_dict[node_dict[subreddit]] = 1
       
            num_words = int(row[6])
            score = int(row[7])

            if (ts in out_dict):
                out_dict[ts][(node_dict[src], node_dict[dst], 0)] = (num_words, score)
                out_dict[ts][(node_dict[src], node_dict[subreddit], 1)] = (num_words, score)
            else:
                out_dict[ts] = {}
                out_dict[ts][(node_dict[src], node_dict[dst], 0)] = (num_words, score)
                out_dict[ts][(node_dict[src], node_dict[subreddit], 1)] = (num_words, score)
            num_lines += 1
    return out_dict, num_lines, node_dict, node_type_dict





def writeNodeType(node_type_dict, outname):
    r"""
    write the node type mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['node_id', 'type'])
        for key in node_type_dict:
            writer.writerow([key, node_type_dict[key]])


def write2csv(outname, out_dict):
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['ts', 'src', 'dst', 'relation_type', 'num_words', 'score'])
        ts_list = list(out_dict.keys())
        ts_list.sort()

        for ts in ts_list:
            for edge in out_dict[ts]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                num_words, score = out_dict[ts][edge]
                row = [ts, head, tail, relation_type, num_words, score]
                writer.writerow(row)


def writeNodeIDMapping(node_dict, outname):
    r"""
    write the node id mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['node_name', 'node_id'])
        for key in node_dict:
            writer.writerow([key, node_dict[key]])

def node_deg_filter(node_deg_dict):
    """
    filter out nodes with degree less than threshold
    """
    deg10_nodes = 0
    deg100_nodes = 0
    deg1000_nodes = 0

    for key in node_deg_dict:
        if (node_deg_dict[key] < 10):
            deg10_nodes += 1
        if (node_deg_dict[key] < 100):
            deg100_nodes += 1
        if (node_deg_dict[key] < 1000):
            deg1000_nodes += 1
    print ("nodes with degree less than 10: ", deg10_nodes)
    print ("nodes with degree less than 100: ", deg100_nodes)
    print ("nodes with degree less than 1000: ", deg1000_nodes)

def find_low_degree_nodes(node_deg_dict, threshold=10):
    """
    find nodes with degree less than threshold
    """
    low_degree_nodes = {}
    for key in node_deg_dict:
        if (node_deg_dict[key] < threshold):
            low_degree_nodes[key] = 1
    return low_degree_nodes


def main():
    fname = "reddit_edgelist.csv"
    _, _, _, _, _, node_deg_dict = load_csv_raw(fname)
    # print ("checking node degree")
    # node_deg_filter(node_deg_dict)
    # print ("checking reddit degree")
    # node_deg_filter(reddit_deg_dict)
    # low_degree_nodes = find_low_degree_nodes(node_deg_dict, threshold=100)
    # save_pkl(low_degree_nodes, 'low_degree_nodes.pkl')


    low_degree_nodes = load_pkl('low_degree_nodes.pkl')
    out_dict, num_lines, node_dict, node_type_dict = load_csv_filtered_node(fname, low_degree_nodes)
    writeNodeType(node_type_dict, 'thgl-forum_nodetype.csv')
    writeNodeIDMapping(node_dict, 'thgl-forum_nodeIDmapping.csv')
    write2csv('thgl-forum_edgelist.csv', out_dict)



if __name__ == "__main__":
    main()


from typing import Optional, Dict, Any, Tuple
import os
import os.path as osp
import numpy as np
import pandas as pd
import shutil
import zipfile
import requests
from clint.textui import progress

from tgb.utils.info import (
    PROJ_DIR,
    DATA_URL_DICT,
    DATA_NUM_CLASSES,
    DATA_VERSION_DICT,
    DATA_EVAL_METRIC_DICT,
    BColors,
)
from tgb.utils.utils import save_pkl, load_pkl
from tgb.utils.pre_process import (
    load_label_dict,
    load_edgelist_sr,
    load_edgelist_token,
    load_edgelist_datetime,
    load_trade_label_dict,
    load_edgelist_trade,
)


class NodePropPredDataset(object):
    def __init__(
        self,
        name: str,
        root: Optional[str] = "datasets",
        meta_dict: Optional[dict] = None,
        preprocess: Optional[bool] = True,
    ) -> None:
        r"""Dataset class for the node property prediction task. Stores meta information about each dataset such as evaluation metrics etc.
        also automatically pre-processes the dataset.
        [!] node property prediction datasets requires the following:
        self.meta_dict["fname"]: path to the edge list file
        self.meta_dict["nodefile"]: path to the node label file

        Parameters:
            name: name of the dataset
            root: root directory to store the dataset folder
            meta_dict: dictionary containing meta information about the dataset, should contain key 'dir_name' which is the name of the dataset folder
            preprocess: whether to pre-process the dataset
        Returns:
            None
        """
        self.name = name  ## original name
        # check if dataset url exist
        if self.name in DATA_URL_DICT:
            self.url = DATA_URL_DICT[self.name]
        else:
            self.url = None
            print(f"Dataset {self.name} url not found, download not supported yet.")

        # check if the evaluatioin metric are specified
        if self.name in DATA_EVAL_METRIC_DICT:
            self.metric = DATA_EVAL_METRIC_DICT[self.name]
        else:
            self.metric = None
            print(
                f"Dataset {self.name} default evaluation metric not found, it is not supported yet."
            )

        root = PROJ_DIR + root

        if meta_dict is None:
            self.dir_name = "_".join(name.split("-"))  ## replace hyphen with underline
            meta_dict = {"dir_name": self.dir_name}
        else:
            self.dir_name = meta_dict["dir_name"]
        self.root = osp.join(root, self.dir_name)
        self.meta_dict = meta_dict
        if "fname" not in self.meta_dict:
            self.meta_dict["fname"] = self.root + "/" + self.name + "_edgelist.csv"
            self.meta_dict["nodefile"] = self.root + "/" + self.name + "_node_labels.csv"

         #! version check
        self.version_passed = True
        self._version_check()

        self._num_classes = DATA_NUM_CLASSES[self.name]

        # initialize
        self._node_feat = None
        self._edge_feat = None
        self._full_data = None
        self.download()
        # check if the root directory exists, if not create it
        if osp.isdir(self.root):
            print("Dataset directory is ", self.root)
        else:
            raise FileNotFoundError(f"Directory not found at {self.root}")

        if preprocess:
            self.pre_process()

        self.label_ts_idx = 0  # index for which node lables to return now

    def _version_check(self) -> None:
        r"""Implement Version checks for dataset files
        updates the file names based on the current version number
        prompt the user to download the new version via self.version_passed variable
        """
        if (self.name in DATA_VERSION_DICT):
            version = DATA_VERSION_DICT[self.name]
        else:
            print(f"Dataset {self.name} version number not found.")
            self.version_passed = False
            return None
        
        if (version > 1):
            #* check if current version is outdated
            self.meta_dict["fname"] = self.root + "/" + self.name + "_edgelist_v" + str(int(version)) + ".csv"
            self.meta_dict["nodefile"] = self.root + "/" + self.name + "_node_labels_v" + str(int(version)) + ".csv"
            
            if (not osp.exists(self.meta_dict["fname"])):
                print(f"Dataset {self.name} version {int(version)} not found.")
                print(f"Please download the latest version of the dataset.")
                self.version_passed = False
                return None

    def download(self) -> None:
        r"""
        downloads this dataset from url
        check if files are already downloaded
        Returns:
            None
        """
        # check if the file already exists
        if osp.exists(self.meta_dict["fname"]) and osp.exists(
            self.meta_dict["nodefile"]
        ):
            print("raw file found, skipping download")
            return

        else:
            inp = input(
                "Will you download the dataset(s) now? (y/N)\n"
            ).lower()  # ask if the user wants to download the dataset
            if inp == "y":
                print(
                    f"{BColors.WARNING}Download started, this might take a while . . . {BColors.ENDC}"
                )
                print(f"Dataset title: {self.name}")

                if self.url is None:
                    raise Exception(
                        "Dataset url not found, download not supported yet."
                    )
                else:
                    r = requests.get(self.url, stream=True)
                    if osp.isdir(self.root):
                        print("Dataset directory is ", self.root)
                    else:
                        os.makedirs(self.root)

                    path_download = self.root + "/" + self.name + ".zip"
                    with open(path_download, "wb") as f:
                        total_length = int(r.headers.get("content-length"))
                        for chunk in progress.bar(
                            r.iter_content(chunk_size=1024),
                            expected_size=(total_length / 1024) + 1,
                        ):
                            if chunk:
                                f.write(chunk)
                                f.flush()
                    # for unzipping the file
                    with zipfile.ZipFile(path_download, "r") as zip_ref:
                        zip_ref.extractall(self.root)
                    print(f"{BColors.OKGREEN}Download completed {BColors.ENDC}")
            else:
                raise Exception(
                    BColors.FAIL
                    + "Data not found error, download "
                    + self.name
                    + " failed"
                )

    def generate_processed_files(
        self,
    ) -> Tuple[pd.DataFrame, Dict[int, Dict[str, Any]]]:
        r"""
        returns an edge list of pandas data frame
        Returns:
            df: pandas data frame storing the temporal edge list
            node_label_dict: dictionary with key as timestamp and item as dictionary of node labels
        """
        OUT_DF = self.root + "/" + "ml_{}.pkl".format(self.name)
        OUT_NODE_DF = self.root + "/" + "ml_{}_node.pkl".format(self.name)
        OUT_LABEL_DF = self.root + "/" + "ml_{}_label.pkl".format(self.name)
        OUT_EDGE_FEAT = self.root + "/" + "ml_{}.pkl".format(self.name + "_edge")

        # * logic for large datasets, as node label file is too big to store on disc
        if self.name == "tgbn-reddit" or self.name == "tgbn-token":
            if osp.exists(OUT_DF) and osp.exists(OUT_NODE_DF) and osp.exists(OUT_EDGE_FEAT):
                df = pd.read_pickle(OUT_DF)
                edge_feat = load_pkl(OUT_EDGE_FEAT)
                if (self.name == "tgbn-token"):
                    #! taking log normalization for numerical stability
                    print ("applying log normalization for weights in tgbn-token")
                    edge_feat[:,0] = np.log(edge_feat[:,0])
                node_ids = load_pkl(OUT_NODE_DF)
                labels_dict = load_pkl(OUT_LABEL_DF)
                node_label_dict = load_label_dict(
                    self.meta_dict["nodefile"], node_ids, labels_dict
                )
                return df, node_label_dict, edge_feat

        # * load the preprocessed file if possible
        if osp.exists(OUT_DF) and osp.exists(OUT_NODE_DF) and osp.exists(OUT_EDGE_FEAT):
            print("loading processed file")
            df = pd.read_pickle(OUT_DF)
            node_label_dict = load_pkl(OUT_NODE_DF)
            edge_feat = load_pkl(OUT_EDGE_FEAT)
        else:  # * process the file
            print("file not processed, generating processed file")
            if self.name == "tgbn-reddit":
                df, edge_feat, node_ids, labels_dict = load_edgelist_sr(
                    self.meta_dict["fname"], label_size=self._num_classes
                )
            elif self.name == "tgbn-token":
                df, edge_feat, node_ids, labels_dict = load_edgelist_token(
                    self.meta_dict["fname"], label_size=self._num_classes
                )
            elif self.name == "tgbn-genre":
                df, edge_feat, node_ids, labels_dict = load_edgelist_datetime(
                    self.meta_dict["fname"], label_size=self._num_classes
                )
            elif self.name == "tgbn-trade":
                df, edge_feat, node_ids = load_edgelist_trade(
                    self.meta_dict["fname"], label_size=self._num_classes
                )

            df.to_pickle(OUT_DF)
            save_pkl(edge_feat, OUT_EDGE_FEAT)

            if self.name == "tgbn-trade":
                node_label_dict = load_trade_label_dict(
                    self.meta_dict["nodefile"], node_ids
                )
            else:
                node_label_dict = load_label_dict(
                    self.meta_dict["nodefile"], node_ids, labels_dict
                )

            if (
                self.name != "tgbn-reddit" and self.name != "tgbn-token"
            ):  # don't save subreddits on disc, the node label file is too big
                save_pkl(node_label_dict, OUT_NODE_DF)
            else:
                save_pkl(node_ids, OUT_NODE_DF)
                save_pkl(labels_dict, OUT_LABEL_DF)
            
            print("file processed and saved")
        return df, node_label_dict, edge_feat

    def pre_process(self) -> None:
        """
        Pre-process the dataset and generates the splits, must be run before dataset properties can be accessed
        Returns:
            None
        """
        # first check if all files exist
        if ("fname" not in self.meta_dict) or ("nodefile" not in self.meta_dict):
            raise Exception("meta_dict does not contain all required filenames")

        df, node_label_dict, edge_feat = self.generate_processed_files()
        sources = np.array(df["u"])
        destinations = np.array(df["i"])
        timestamps = np.array(df["ts"])
        edge_idxs = np.array(df["idx"])
        edge_label = np.ones(sources.shape[0])
        #self._edge_feat = np.array(df["w"])
        self._edge_feat = edge_feat

        full_data = {
            "sources": sources,
            "destinations": destinations,
            "timestamps": timestamps,
            "edge_idxs": edge_idxs,
            "edge_feat": self._edge_feat,
            "edge_label": edge_label,
            "node_label_dict": node_label_dict,
        }
        self._full_data = full_data

        # storing the split masks
        _train_mask, _val_mask, _test_mask = self.generate_splits(full_data)

        self._train_mask = _train_mask
        self._val_mask = _val_mask
        self._test_mask = _test_mask

        self.label_dict = node_label_dict
        self.label_ts = np.array(list(node_label_dict.keys()))
        self.label_ts = np.sort(self.label_ts)

    def generate_splits(
        self,
        full_data: Dict[str, Any],
        val_ratio: float = 0.15,
        test_ratio: float = 0.15,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""
        Generates train, validation, and test splits from the full dataset
        Parameters:
            full_data: dictionary containing the full dataset
            val_ratio: ratio of validation data
            test_ratio: ratio of test data
        Returns:
            train_mask: boolean mask for training data
            val_mask: boolean mask for validation data
            test_mask: boolean mask for test data
        """
        val_time, test_time = list(
            np.quantile(
                full_data["timestamps"],
                [(1 - val_ratio - test_ratio), (1 - test_ratio)],
            )
        )
        timestamps = full_data["timestamps"]
        train_mask = timestamps <= val_time
        val_mask = np.logical_and(timestamps <= test_time, timestamps > val_time)
        test_mask = timestamps > test_time

        return train_mask, val_mask, test_mask

    def find_next_labels_batch(
        self,
        cur_t: int,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        r"""
        this returns the node labels closest to cur_t (for that given day)
        Parameters:
            cur_t: current timestamp of the batch of edges
        Returns:
            ts: timestamp of the node labels
            source_idx: node ids
            labels: the stacked label vectors
        """
        if self.label_ts_idx >= (self.label_ts.shape[0]):
            # for query that are after the last batch of labels
            return None
        else:
            ts = self.label_ts[self.label_ts_idx]

        if cur_t >= ts:
            self.label_ts_idx += 1  # move to the next ts
            # {ts: {node_id: label_vec}}
            node_ids = np.array(list(self.label_dict[ts].keys()))

            node_labels = []
            for key in self.label_dict[ts]:
                node_labels.append(np.array(self.label_dict[ts][key]))
            node_labels = np.stack(node_labels, axis=0)
            label_ts = np.full(node_ids.shape[0], ts, dtype="int")
            return (label_ts, node_ids, node_labels)
        else:
            return None

    def reset_label_time(self) -> None:
        r"""
        reset the pointer for node label once the entire dataset has been iterated once
        Returns:
            None
        """
        self.label_ts_idx = 0

    def return_label_ts(self) -> int:
        """
        return the current label timestamp that the pointer is at
        Returns:
            ts: int, the timestamp of the node labels
        """
        if (self.label_ts_idx >= self.label_ts.shape[0]):
            return self.label_ts[-1]
        else:
            return self.label_ts[self.label_ts_idx]

    @property
    def num_classes(self) -> int:
        """
        number of classes in the node label
        Returns:
            num_classes: int, number of classes
        """
        return self._num_classes

    @property
    def eval_metric(self) -> str:
        """
        the official evaluation metric for the dataset, loaded from info.py
        Returns:
            eval_metric: str, the evaluation metric
        """
        return self.metric

    # TODO not sure needed, to be removed
    @property
    def node_feat(self) -> Optional[np.ndarray]:
        r"""
        Returns the node features of the dataset with dim [N, feat_dim]
        Returns:
            node_feat: np.ndarray, [N, feat_dim] or None if there is no node feature
        """
        return self._node_feat

    # TODO not sure needed, to be removed
    @property
    def edge_feat(self) -> Optional[np.ndarray]:
        r"""
        Returns the edge features of the dataset with dim [E, feat_dim]
        Returns:
            edge_feat: np.ndarray, [E, feat_dim] or None if there is no edge feature
        """
        return self._edge_feat
    
    @property
    def node_label_dict(self) -> Dict[int, Dict[int, Any]]:
        r"""
        Returns the node label dictionary of the dataset with {timestamp: {node_id: label_vec}}
        Returns:
            label_dict: Dict[int, Dict[int, Any]], the node label dictionary
        """
        return self.label_dict

    @property
    def full_data(self) -> Dict[str, Any]:
        r"""
        the full data of the dataset as a dictionary with keys: 'sources', 'destinations', 'timestamps', 'edge_idxs', 'edge_feat', 'w', 'edge_label',

        Returns:
            full_data: Dict[str, Any]
        """
        if self._full_data is None:
            raise ValueError(
                "dataset has not been processed yet, please call pre_process() first"
            )
        return self._full_data

    @property
    def train_mask(self) -> np.ndarray:
        r"""
        Returns the train mask of the dataset
        Returns:
            train_mask
        """
        if self._train_mask is None:
            raise ValueError("training split hasn't been loaded")
        return self._train_mask

    @property
    def val_mask(self) -> np.ndarray:
        r"""
        Returns the validation mask of the dataset
        Returns:
            val_mask: Dict[str, Any]
        """
        if self._val_mask is None:
            raise ValueError("validation split hasn't been loaded")

        return self._val_mask

    @property
    def test_mask(self) -> np.ndarray:
        r"""
        Returns the test mask of the dataset:
        Returns:
            test_mask: Dict[str, Any]
        """
        if self._test_mask is None:
            raise ValueError("test split hasn't been loaded")

        return self._test_mask


def main():
    # download files
    name = "tgbn-trade" 
    dataset = NodePropPredDataset(name=name, root="datasets", preprocess=True)

    dataset.node_feat
    dataset.edge_feat  # not the edge weights
    dataset.full_data
    dataset.full_data["edge_idxs"]
    dataset.full_data["sources"]
    dataset.full_data["destinations"]
    dataset.full_data["timestamps"]
    dataset.full_data["y"]

    train_data = dataset.full_data[dataset.train_mask]
    val_data = dataset.full_data[dataset.val_mask]
    test_data = dataset.full_data[dataset.test_mask]


if __name__ == "__main__":
    main()



import csv
import datetime
import glob, os



def main():
    train_fname = "train.txt"
    val_fname = "valid.txt"
    test_fname = "test.txt"
    
    train_dict, num_lines = load_csv(train_fname)
    print ("there are ", num_lines, " lines in the train file")
    print ("there are ", len(train_dict), " timestamps in the train file")
    val_dict, num_lines = load_csv(val_fname)
    print ("there are ", num_lines, " lines in the val file")
    print ("there are ", len(val_dict), " timestamps in the val file")
    test_dict, num_lines = load_csv(test_fname)
    print ("there are ", num_lines, " lines in the test file")
    print ("there are ", len(test_dict), " timestamps in the test file")

    train_dict.update(val_dict)
    train_dict.update(test_dict)
    print ("there are ", len(train_dict), " timestamps in the combined file")

    outname = "tkgl-yago_edgelist.csv"
    write_csv(outname, train_dict)


def write_csv(outname, out_dict):
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['timestamp', 'head', 'tail', 'relation_type'])
        for ts in out_dict:
            for edge in out_dict[ts]:
                src = edge[0]
                rel_type = edge[1]
                dst = edge[2]
                row = [ts, src, dst, rel_type]
                writer.writerow(row)


def load_csv(fname):
    out_dict = {}
    num_lines = 0
    with open(fname, 'r') as f:
        reader = csv.reader(f, delimiter ='\t')
        #! src rel_type dst ts 
        # 10289	9	10290	0	0
        for row in reader: 
            src = int (row[0])
            rel_type = int (row[1])
            dst = int (row[2])
            ts = int (row[3])
            if ts not in out_dict:
                out_dict[ts] = {(src,rel_type,dst):1}
            else:
                out_dict[ts][(src,rel_type,dst)] = 1
            num_lines += 1
    return out_dict, num_lines



if __name__ == "__main__":
    main()


import csv
import datetime

def count_node_freq(fname, filter_size=100):

    node_dict = {}
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        ctr = 0
        for row in csv_reader:
            if ctr == 0:
                ctr += 1
                continue
            else:
                token_type = row[0]
                src = row[1]
                if (src not in node_dict):
                    node_dict[src] = 1
                else:
                    node_dict[src] += 1
                dst = row[2]
                if (dst not in node_dict):
                    node_dict[dst] = 1
                else:
                    node_dict[dst] += 1
                ctr += 1

    num_10 = 0
    num_100 = 0
    num_1000 = 0
    num_2000 = 0
    num_5000 = 0 

    for node in node_dict:
        if node_dict[node] >= 10:
            num_10 += 1
        if node_dict[node] >= 100:
            num_100 += 1
        if node_dict[node] >= 1000:
            num_1000 += 1
        if node_dict[node] >= 2000:
            num_2000 += 1
        if node_dict[node] >= 5000:
            num_5000 += 1

    print("number of nodes with # edges >= 10 is ", num_10)
    print("number of nodes with # edges >= 100 is ", num_100)
    print("number of nodes with # edges >= 1000 is ", num_1000)
    print("number of nodes with # edges >= 2000 is ", num_2000)
    print("number of nodes with # edges >= 5000 is ", num_5000)
    print("----------------------high level statistics-------------------------")


    #! keep nodes with at least 100 edges
    node_dict_filtered = {}
    for node in node_dict:
        if node_dict[node] >= filter_size:
            node_dict_filtered[node] = node_dict[node]
    return node_dict_filtered






def filter_edgelist(token_fname, edgefile, outname):
    """
    preserve only the tokens in the token file
    Parameters:
        token_fname: the file of the token file
        edgefile: the edgelist file name
        outname: the output filtered edgelistname
    """
    #* read tokens from the file
    token_dict = {}
    with open(token_fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        ctr = 0
        for row in csv_reader:
            if ctr == 0:
                ctr += 1
                continue
            else:
                token_type = row[0]
                token_dict[token_type] = 1
    
    with open(edgefile, "r") as in_file:
        with open(outname, "w") as out_file:
            csv_reader = csv.reader(in_file, delimiter=",")
            csv_writer = csv.writer(out_file, delimiter=",")
            csv_writer.writerow(["token_address", "from_address", "to_address", "value", "block_timestamp"])
            ctr = 0
            for row in csv_reader:
                if ctr == 0:
                    ctr += 1
                    continue
                else:
                    token_type = row[0]
                    if token_type in token_dict:
                        csv_writer.writerow(row)
                    ctr += 1


def filter_by_node(node_dict, edgefile, outname):
    with open(edgefile, "r") as in_file:
        with open(outname, "w") as out_file:
            csv_reader = csv.reader(in_file, delimiter=",")
            csv_writer = csv.writer(out_file, delimiter=",")
            csv_writer.writerow(["token_address", "from_address", "to_address", "value", "block_timestamp"])
            ctr = 0
            for row in csv_reader:
                if ctr == 0:
                    ctr += 1
                    continue
                else:
                    token_type = row[0]
                    src = row[1]
                    dst = row[2]
                    if (src in node_dict) or (dst in node_dict):
                        csv_writer.writerow(row)
                    ctr += 1



def store_node_list(node_dict, outname):
    """
    Parameters:
        outname: name of the output csv file
    Output:
        output csv file with node list
    """
    with open(outname, "w") as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=",")
        csv_writer.writerow(["node_list", "frequency"])
        for key, value in node_dict.items():
            csv_writer.writerow([key, value])


def load_node_dict(fname):
    node_dict = {}
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        ctr = 0
        for row in csv_reader:
            if ctr == 0:
                ctr += 1
                continue
            else:
                node = row[0]
                freq = int(row[1])
                node_dict[node] = freq
                ctr += 1
    return node_dict







def store_token_address(token_dict, outname, topk=1000):
    """
    Parameters:
        outname: name of the output csv file
    Output:
        output csv file with topk token addresses
    """
    sorted_tokens = {k: v for k, v in sorted(token_dict.items(), key=lambda item: item[1], reverse=True)}
    ctr = 0
    with open(outname, "w") as csv_file:
        csv_writer = csv.writer(csv_file, delimiter=",")
        csv_writer.writerow(["token_address", "frequency"])
        for key, value in sorted_tokens.items():
            if (ctr <= topk):
                csv_writer.writerow([key, value])
            else:
                break
            ctr += 1

def analyze_token_frequency(fname):
    # ['token_address', 'from_address', 'to_address', 'value', 'block_timestamp']
    token_dict = {}
    node_dict = {}
    time_dict = {}
    max_w = 0
    min_w = 100000
    num_edges = 0

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        ctr = 0
        for row in csv_reader:
            if ctr == 0:
                ctr += 1
                continue
            else:
                token_type = row[0]
                if (token_type not in token_dict):
                    token_dict[token_type] = 1
                else:
                    token_dict[token_type] += 1
                src = row[1]
                if (src not in node_dict):
                    node_dict[src] = 1
                else:
                    node_dict[src] += 1
                dst = row[2]
                if (dst not in node_dict):
                    node_dict[dst] = 1
                else:
                    node_dict[dst] += 1

                w = float(row[3])
                if (w > max_w):
                    max_w = w
                elif (w < min_w):
                    min_w = w
                timestamp = row[4]
                if (timestamp not in time_dict):
                    time_dict[timestamp] = 1
                ctr += 1
                num_edges += 1

    print ( "number of edges are ", num_edges)
    print (" number of unique tokens are ", len(token_dict))
    print (" number of unique nodes are ", len(node_dict))
    print (" number of unique timestamps are ", len(time_dict))
    print (" max weight is ", max_w)
    print (" min weight is ", min_w)

    # topk = 1000
    # store_token_address(token_dict, "token_list.csv", topk=topk)

def to_bipartite(in_name, out_name, node_dict):
    """
    load and convert a user-user graph into a user-token bipartite graph
    """
    with open(in_name, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        with open(out_name, "w") as out_file:
            csv_writer = csv.writer(out_file, delimiter=",")
            csv_writer.writerow(["timestamp", "user_address", "token_address", "value", "IsSender"])
            ctr = 0
            for row in csv_reader:
                if ctr == 0:
                    ctr += 1
                    continue
                else:
                    token_type = row[0]
                    src = row[1]
                    dst = row[2]
                    w = float(row[3])
                    timestamp = row[4]
                    if (src in node_dict):
                        csv_writer.writerow([timestamp, src, token_type, w, 1])
                    if (dst in node_dict):
                        csv_writer.writerow([timestamp, dst, token_type, w, 0])
                    

def analyze_csv(fname):
    node_dict = {}
    edge_dict = {}
    num_edges = 0
    num_time = 0
    time_dict = {}

    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        line_count = 0
        for row in csv_reader:
            if line_count == 0:
                line_count += 1
            else:
                # t,u,v,w
                t = row[0]
                u = row[1]
                v = row[2]

                # count unique time
                if t not in time_dict:
                    time_dict[t] = 1
                    num_time += 1

                # unique nodes
                if u not in node_dict:
                    node_dict[u] = 1
                else:
                    node_dict[u] += 1

                if v not in node_dict:
                    node_dict[v] = 1
                else:
                    node_dict[v] += 1

                # unique edges
                num_edges += 1
                if (u, v) not in edge_dict:
                    edge_dict[(u, v)] = 1
                else:
                    edge_dict[(u, v)] += 1

    print("----------------------high level statistics-------------------------")
    print("number of total edges are ", num_edges)
    print("number of nodes are ", len(node_dict))
    print("number of unique edges are ", len(edge_dict))
    print("number of unique timestamps are ", num_time)

    num_10 = 0
    num_100 = 0
    num_1000 = 0

    for node in node_dict:
        if node_dict[node] >= 10:
            num_10 += 1
        if node_dict[node] >= 100:
            num_100 += 1
        if node_dict[node] >= 1000:
            num_1000 += 1
    print("number of nodes with # edges >= 10 is ", num_10)
    print("number of nodes with # edges >= 100 is ", num_100)
    print("number of nodes with # edges >= 1000 is ", num_1000)
    print("----------------------high level statistics-------------------------")        



def convert_2_sec(fname, outname):
    """
    convert datetime object format = "%Y-%m-%d %H:%M:%S" to seconds
    #2017-07-24 17:48:15+00:00
    """
    format = "%Y-%m-%d %H:%M:%S"
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        with open(outname, "w") as out_file:
            csv_writer = csv.writer(out_file, delimiter=",")
            csv_writer.writerow(["timestamp", "user_address", "token_address", "value", "IsSender"])
            ctr = 0
            for row in csv_reader:
                if ctr == 0:
                    ctr += 1
                    continue
                else:
                    timestamp = row[0][:19]
                    date_object = datetime.datetime.strptime(timestamp, format)
                    timestamp_sec = int(date_object.timestamp())
                    src = row[1]
                    dst = row[2]
                    w = float(row[3])
                    IsSender = int(row[4])
                    if (w != 0):
                        csv_writer.writerow([timestamp_sec, src, dst, w, IsSender])

    



def print_csv(fname):
    # ['token_address', 'from_address', 'to_address', 'value', 'block_timestamp']
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        ctr = 0
        for row in csv_reader:
            ctr += 1
    print ("there are ", ctr, " rows in the csv file")



def sort_edgelist_by_time(fname, outname):
    row_dict = {}
    with open(fname, "r") as csv_file:
        csv_reader = csv.reader(csv_file, delimiter=",")
        with open(outname, "w") as out_file:
            csv_writer = csv.writer(out_file, delimiter=",")
            csv_writer.writerow(["timestamp", "user_address", "token_address", "value", "IsSender"])
            ctr = 0
            for row in csv_reader:
                if ctr == 0:
                    ctr += 1
                    continue
                else:
                    timestamp =int(row[0])
                    if (timestamp not in row_dict):
                        row_dict[timestamp] = [row]
                    else:
                        row_dict[timestamp].append(row)
            for i in sorted(row_dict.keys()):
                rows = row_dict[i]
                for row in rows:
                    csv_writer.writerow(row)





#! aggregate node labels
def generate_aggregate_labels(fname: str, outname: str, days: int = 7):
    """
    aggregate the genres over a number of days,  as specified by days
    prediction should always be at the first second of the day
    #! daily labels are always shifted by 1 day
    """

    ts_prev = 0

    DAY_IN_SEC = 86400
    timespan = days * DAY_IN_SEC

    user_dict = {}

    # ts, src, subreddit, num_words, score
    with open(outname, "w") as outf:
        write = csv.writer(outf)
        fields = ["ts", "user_address", "token_address", "weight"] #["ts", "user", "subreddit", "weight"]
        write.writerow(fields)

        with open(fname, "r") as csv_file:
            csv_reader = csv.reader(csv_file, delimiter=",")
            line_count = 0
            # ts, src, subreddit, num_words, score
            for row in csv_reader:
                if line_count == 0:
                    line_count += 1
                else:
                    ts = float(row[0])
                    ts = int(ts)
                    user = row[1]
                    item = row[2]
                    w = float(row[3])
                    if (w == 0):
                        print (row)

                    if line_count == 1:
                        ts_prev = ts

                    if (ts - ts_prev) > timespan:
                        for user in user_dict:
                            total = sum(user_dict[user].values())
                            item_dict = {
                                k: v / total for k, v in user_dict[user].items()
                            }
                            for item, w in item_dict.items():
                                write.writerow(
                                    [ts_prev + DAY_IN_SEC, user, item, w]
                                )
                        user_dict = {}
                        ts_prev = ts_prev + DAY_IN_SEC  #! move label to the next day
                    else:
                        if user in user_dict:
                            if item in user_dict[user]:
                                user_dict[user][item] += w
                            else:
                                user_dict[user][item] = w
                        else:
                            user_dict[user] = {}
                            user_dict[user][item] = w
                    line_count += 1




def main():

    """
    processing token types
    """
    # fname = "ERC20_token_network.csv"
    # #analyze_token_frequency(fname)

    # token_file = "token_list.csv"
    # outname = "filtered_token_edgelist.csv"

    #! filter by token frequency
    # filter_edgelist(token_file, fname, outname)
    # #print_csv(fname)
    # #analyze_csv(fname)

    """
    processing node dict
    """
    # fname = "filtered_token_edgelist.csv"
    # #! filter by node frequency
    # node_dict = count_node_freq(fname, filter_size=100)
    # store_node_list(node_dict, "node_list.csv")
    # #store_token_address(node_dict, "node_list.csv", topk=0)

    # outname = "tgbl-token-edgelist_100.csv"
    # filter_by_node(node_dict, fname, outname)
    # analyze_token_frequency('tgbl-token-edgelist_100.csv')


    #! converting user-user graph to user-token bipartite graph
    # out_name = "tgbl-token_edgelist.csv"
    # node_dict = load_node_dict("node_list.csv")
    # to_bipartite('tgbl-token-edgelist_100.csv', out_name, node_dict)
    # analyze_csv(out_name)


    #! convert datetime to seconds
    #convert_2_sec("tgbl-token_edgelist_old.csv", "tgbn-token_edgelist.csv")


    #! sort the timestamps in the edgelist
    # fname = "tgbn-token_edgelist.csv"
    # outname = "tgbn-token_edgelist_sorted.csv"
    # sort_edgelist_by_time(fname, outname)



    #! generate node labels
    edgefile = "tgbn-token_edgelist.csv"
    outfile = "tgbn-token_node_labels.csv"
    days = 7
    generate_aggregate_labels(edgefile, outfile, days=days)





    




if __name__ == "__main__":
    main()


import csv
import datetime
import glob, os




def load_csv_raw(fname):
    r"""
    load from the raw data and retrieve, timestamp, head, tail, relation 
    convert all dates into unix timestamps
    #!Event ID	Event Date	Event Type	Event Mode	Intensity	Quad Code	Contexts	Actor Name	Actor Country	Actor COW	Primary Actor Sector	Actor Sectors	Actor Title	Actor Name Raw	Wikipedia Actor ID	Recipient Name	Recipient Country	Recipient COW	Primary Recipient Sector	Recipient Sectors	Recipient Title	Recipient Name Raw	Wikipedia Recipient ID	Placename	City	District	Province	Country	Latitude	Longitude	GeoNames ID	Raw Placename	Feature Type	Source	Publication Date	Story People	Story Organizations	Story Locations	Language	Version
    """
    out_dict = {}
    first_row = True
    num_lines = 0
    with open(fname, 'r', encoding='ISO-8859-1') as f:
        reader = csv.reader(f, delimiter ='\t')
        for row in reader: 
            if first_row:
                first_row = False
                continue
            date = row[1]
            relation_type = row[2]
            head = row[7]
            tail = row[15]

            if (len(date) == 0):
                continue
            
            if ("None" in date or "None" in head or "None" in tail or "None" in relation_type):
                continue
            else:
                 #! remove redundant edges with same timestamps
                TIME_FORMAT = "%Y-%m-%d" #2018-01-01
                date_cur = datetime.datetime.strptime(date, TIME_FORMAT)
                ts = int(date_cur.timestamp())
                num_lines += 1
                if (ts in out_dict):
                    if (head, tail, relation_type) in out_dict[ts]:
                        out_dict[ts][(head, tail, relation_type)] += 1
                    else:
                        out_dict[ts][(head, tail, relation_type)] = 1
                else:
                    out_dict[ts] = {}
                    out_dict[ts][(head, tail, relation_type)] = 1
    return out_dict, num_lines


#! fill in node and edge type dictionaries
def write2csv(outname: str, 
              out_dict: dict,
              edge_type_dict: dict = None,
              node_dict: dict = None,):
    r"""
    Write the dictionary to a csv file
    also keep track of edge_type or node_dict, update the provided one too
    """
    if (edge_type_dict is None):
        edge_type_dict = {}
    if (node_dict is None):
        node_dict = {}

    max_edge_type_id = len(edge_type_dict)
    max_node_id = len(node_dict)    
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['date', 'head', 'tail', 'relation_type'])

        dates = list(out_dict.keys())
        dates.sort()
        for date in dates:
            for edge in out_dict[date]:
                head = edge[0]
                tail = edge[1]
                relation_type = edge[2]
                if head not in node_dict:
                    node_dict[head] = max_node_id
                    max_node_id += 1
                if tail not in node_dict:
                    node_dict[tail] = max_node_id
                    max_node_id += 1
                if relation_type not in edge_type_dict:
                    edge_type_dict[relation_type] = max_edge_type_id
                    max_edge_type_id += 1
                row = [date, node_dict[head], node_dict[tail], edge_type_dict[relation_type]]
                writer.writerow(row)
    
    return edge_type_dict, node_dict


# def write2csv(outname, out_dict):

#     node_dict = {}
#     max_node_id = 0
#     edge_type_dict = {}
#     max_edge_type_id = 0

#     with open(outname, 'w') as f:
#         writer = csv.writer(f, delimiter =',')
#         writer.writerow(['date', 'head', 'tail', 'relation_type'])

#         for date in out_dict:
#             for edge in out_dict[date]:
#                 head = edge[0]
#                 tail = edge[1]
#                 relation_type = edge[2]
#                 if head not in node_dict:
#                     node_dict[head] = max_node_id
#                     max_node_id += 1
#                 if tail not in node_dict:
#                     node_dict[tail] = max_node_id
#                     max_node_id += 1
#                 if relation_type not in edge_type_dict:
#                     edge_type_dict[relation_type] = max_edge_type_id
#                     max_edge_type_id += 1
#                 row = [date, node_dict[head], node_dict[tail], edge_type_dict[relation_type]]
#                 writer.writerow(row)


def writeEdgeTypeMapping(edge_type_dict, outname):
    r"""
    write the edge type mapping to a file
    """
    with open(outname, 'w') as f:
        writer = csv.writer(f, delimiter =',')
        writer.writerow(['edge_id', 'type'])
        for key in edge_type_dict:
            writer.writerow([key, edge_type_dict[key]])


def main():

    #example
    # fname = "2018-Jan.txt"
    # print ("hi")
    # lines = load_csv_raw(fname)
    # outname = "tkgl-polecat_edgelist.csv"
    # write2csv(outname, lines)

    total_lines = 0
    num_days = 0
    total_edge_dict = {}

    #1. find all files with .txt in the folder
    for file in glob.glob("*.csv"):
        outname = file[0:7] + "_edgelist.csv"
        print ("processing", file, "to", outname)
        edge_dict, num_lines = load_csv_raw(file)
        total_lines += num_lines
        num_days += len(edge_dict)
        total_edge_dict.update(edge_dict)
    edge_type_dict, node_dict = write2csv("tkgl-polecat_edgelist.csv", total_edge_dict)
    print ("-----------------------------------")
    print ("total number of lines", total_lines)
    print ("total number of days", num_days)
    print ("there are", len(edge_type_dict), "unique edge types")
    print ("there are", len(node_dict), "unique nodes")
    writeEdgeTypeMapping(edge_type_dict, "tkgl-polecat_edgemapping.csv")



if __name__ == "__main__":
    main()

    #* rename functions
    # renames = []
    # for file in glob.glob("*.txt"):
    #     outname = file[-12:-4] + "_edgelist.csv"
    #     file_rename = file[-12:-4] + "_raw.csv"
    #     if ("Jan" in outname):
    #         outname = outname.replace("Jan", "01")
    #         renames.append((file, file_rename.replace("Jan", "01")))
    #     elif ("Feb" in outname):
    #         outname = outname.replace("Feb", "02")
    #         renames.append((file, file_rename.replace("Feb", "02")))
    #     elif ("Mar" in outname):
    #         outname = outname.replace("Mar", "03")
    #         renames.append((file, file_rename.replace("Mar", "03")))
    #     elif ("Apr" in outname):
    #         outname = outname.replace("Apr", "04")
    #         renames.append((file, file_rename.replace("Apr", "04")))
    #     elif ("May" in outname):
    #         outname = outname.replace("May", "05")
    #         renames.append((file, file_rename.replace("May", "05")))
    #     elif ("Jun" in outname):
    #         outname = outname.replace("Jun", "06")
    #         renames.append((file, file_rename.replace("Jun", "06")))
    #     elif ("Jul" in outname):
    #         outname = outname.replace("Jul", "07")
    #         renames.append((file, file_rename.replace("Jul", "07")))
    #     elif ("Aug" in outname):
    #         outname = outname.replace("Aug", "08")
    #         renames.append((file, file_rename.replace("Aug", "08")))
    #     elif ("Sep" in outname):
    #         outname = outname.replace("Sep", "09")
    #         renames.append((file, file_rename.replace("Sep", "09")))
    #     elif ("Oct" in outname):
    #         outname = outname.replace("Oct", "10")
    #         renames.append((file, file_rename.replace("Oct", "10")))
    #     elif ("Nov" in outname):
    #         outname = outname.replace("Nov", "11")
    #         renames.append((file, file_rename.replace("Nov", "11")))
    #     elif ("Dec" in outname):
    #         outname = outname.replace("Dec", "12")
    #         renames.append((file, file_rename.replace("Dec", "12")))
    # for file, file_rename in renames:
    #     os.rename(file, file_rename)


"""
Dataset statistics
"""

import numpy as np
import pandas as pd
import networkx as nx
import argparse

from torch_geometric.loader import TemporalDataLoader
from tgb.nodeproppred.dataset_pyg import PyGNodePropPredDataset
# from tgb.linkproppred.dataset_pyg import PyGLinkPropPredDataset
from tgb.linkproppred.dataset import LinkPropPredDataset




def get_unique_edges(sources, destination):
    r"""
    return unique edges
    """
    unique_e = {}
    for src, dst in zip(sources, destination):
        if (src, dst) not in unique_e:
            unique_e[(src, dst)] = True
    return unique_e


def get_avg_e_per_ts(edgelist_df):
    r"""
    get the average number of edges per each timestamp
    """
    sum_num_e_per_ts = 0
    unique_ts = np.unique(np.array(edgelist_df['ts'].tolist()))
    for ts in unique_ts:
        num_e_at_this_ts = len(edgelist_df.loc[edgelist_df['ts'] == ts])
        sum_num_e_per_ts += num_e_at_this_ts
    avg_num_e_per_ts = (sum_num_e_per_ts * 1.0) / len(unique_ts)
    
    # print(f"INFO: avg_num_e_per_ts: {avg_num_e_per_ts}")
    return avg_num_e_per_ts


def get_avg_degree(edgelist_df):
    r"""
    get average degree over the timestamps
    """
    degree_avg_at_ts_list = []
    unique_ts = np.unique(np.array(edgelist_df['ts'].tolist()))
    for ts in unique_ts:  
        e_at_this_ts = edgelist_df.loc[edgelist_df['ts'] == ts]
        G = nx.MultiGraph()
        for idx, e_row in e_at_this_ts.iterrows():
            G.add_edge(e_row['src'], e_row['dst'], weight=e_row['ts'])
        nodes = G.nodes()
        degrees = [G.degree[n] for n in nodes]
        degree_avg_at_ts_list.append(np.mean(degrees))

    # print(f"INFO: avg_degree: {np.mean(degree_avg_at_ts_list)}")
    
    return np.mean(degree_avg_at_ts_list)


def get_index_metrics(train_val_data, test_data):
    r"""
    compute `surprise` and `recurrence` indices
    """
    train_val_e_set = {}
    for src, dst in zip(train_val_data['sources'], train_val_data['destinations']):
        if (src, dst) not in train_val_e_set:
            train_val_e_set[(src, dst)] = True
    
    test_e_set = {}
    for src, dst in zip(test_data['sources'], test_data['destinations']):
        if (src, dst) not in test_e_set:
            test_e_set[(src, dst)] = True
    
    train_val_size = len(train_val_data['sources'])
    test_size = len(test_data['sources'])

    intersect = difference = 0
    for e in test_e_set:
        if e in train_val_e_set:
            intersect += 1
        else:
            difference += 1

    surprise = float(difference * 1.0 / test_size)
    reoccurrence = float(intersect * 1.0 / train_val_size)
    return surprise, reoccurrence


def get_node_ratio(history_data, eval_data):
    r"""
    compute the ratio of new nodes
    """
    eval_uniq_nodes = set(eval_data['sources']).union(set(eval_data['destinations'])) 
    hist_uniq_nodes = set(history_data['sources']).union(set(history_data['destinations'])) 
    new_nodes = []
    for node in eval_uniq_nodes:
        if node not in hist_uniq_nodes:
            new_nodes.append(node)
    new_nodes = set(new_nodes)
    new_node_ratio = float(len(new_nodes) * 1.0 / len(eval_uniq_nodes))

    return new_node_ratio


def get_dataset_stats(data, temporal_stats=False):
    r"""
    returns simple stats based on counts
    """
    # simple stats
    sources, destinations, timestamps = data['full']['sources'], data['full']['destinations'], data['full']['timestamps']
    edgelist_df = pd.DataFrame(zip(sources, destinations, timestamps), columns=['src', 'dst', 'ts'])
    num_nodes = len(np.unique(np.concatenate((sources, destinations), axis=0)))
    num_edges = len(sources)  # = len(destinations) = len(timestamps)
    num_unique_ts = len(np.unique(timestamps))
    unique_e = get_unique_edges(sources, destinations)
    num_unique_e = len(unique_e)

    # compute temporal stats
    if temporal_stats:  # because it takes so long for large datasets...
        avg_e_per_ts = get_avg_e_per_ts(edgelist_df)
        avg_degree_per_ts = get_avg_degree(edgelist_df)
    else:
        avg_e_per_ts = -1
        avg_degree_per_ts = -1
    
    # compute reoccurrence & surprise
    surprise, reoccurrence = get_index_metrics(data['train_val'], data['test'])

    # compute new node ratio 
    val_nn_ratio = get_node_ratio(data['train'], data['val'])
    #test_nn_ratio = get_node_ratio(data['train_val'], data['test'])
    test_nn_ratio = get_node_ratio(data['train'], data['test'])


    stats_dict = {
                  'num_nodes': num_nodes,
                  'num_edges': num_edges,
                  'num_unique_ts': num_unique_ts,
                  'num_unique_e': num_unique_e,
                  'avg_e_per_ts': avg_e_per_ts,
                  'avg_degree_per_ts': avg_degree_per_ts,
                  'surprise': surprise,
                  'reocurrence': reoccurrence,
                  'val_nn_ratio': val_nn_ratio,
                  'test_nn_ratio': test_nn_ratio,
                  }
    return stats_dict


def main():
    r"""
    Generate dateset statistics
    """
    parser = argparse.ArgumentParser(description='Dataset statistics')
    parser.add_argument('-d', '--data', type=str, default='tgbl-wiki', help='random seed to use')
    parser.add_argument('--tempstats', action='store_true', default=False, help='whether compute temporal statistics')
    parser.parse_args()
    args = parser.parse_args()

    DATA = args.data
    temporal_stats = args.tempstats

    # data loading ...
    if DATA in ['tgbl-wiki', 'tgbl-review', 'tgbl-flight', 'tgbl-comment', 'tgbl-coin']:
        # load data: link prop. pred. with `numpy`
        dataset = LinkPropPredDataset(name=DATA, root="datasets", preprocess=True)
        data = dataset.full_data  

        # get masks
        train_mask = dataset.train_mask
        val_mask = dataset.val_mask
        test_mask = dataset.test_mask
        train_data = {'sources': data['sources'][train_mask],
                      'destinations': data['destinations'][train_mask],
                      }
        val_data = {'sources': data['sources'][val_mask],
                      'destinations': data['destinations'][val_mask],
                      }
        train_val_data = {'sources': np.concatenate([data['sources'][train_mask], data['sources'][val_mask]]),
                      'destinations': np.concatenate([data['destinations'][train_mask], data['destinations'][val_mask]]),
                      }
        test_data = {'sources': data['sources'][test_mask],
                      'destinations': data['destinations'][test_mask],
                      }
        full_data = {'sources': data['sources'], 
                     'destinations': data['destinations'], 
                     'timestamps': data['timestamps'],
                     }

    elif DATA in ['tgbn-trade', 'tgbn-genre', 'tgbn-reddit', 'tgbn-token']:
        # load data: node prop. pred.
        dataset = PyGNodePropPredDataset(name=DATA, root="datasets")
        data = dataset.get_TemporalData()
        
        # split data
        train_mask = dataset.train_mask
        val_mask = dataset.val_mask
        test_mask = dataset.test_mask
        train_val_mask = np.logical_or(np.array(train_mask), np.array(val_mask))

        train_data = {'sources': np.array(data[train_mask].src),
                      'destinations': np.array(data[train_mask].dst),
                      }
        val_data = {'sources': np.array(data[val_mask].src),
                    'destinations': np.array(data[val_mask].dst),
                    }
        train_val_data = {'sources': np.concatenate([np.array(data[train_mask].src), np.array(data[val_mask].src)]),
                          'destinations': np.concatenate([np.array(data[train_mask].dst), np.array(data[val_mask].dst)]),
                          }
        test_data = {'sources': np.array(data[test_mask].src),
                     'destinations': np.array(data[test_mask].dst),
                     } 
        full_data = {'sources': np.array(data.src), 
                     'destinations': np.array(data.dst), 
                     'timestamps': np.array(data.t),
                     }

    else:
        raise ValueError("Unsupported data!")

    split_data = {'train': train_data,
                  'val': val_data,
                  'train_val': train_val_data,
                  'test': test_data,
                  'full': full_data,
                  }
    # compute dataset statistics...
    print("=============================")
    print(f">>> DATA: {DATA}")
    dataset_stats = get_dataset_stats(split_data, temporal_stats)
    for k, v in dataset_stats.items():
        print(f"{k}: {v}")
    print("=============================")


if __name__ == "__main__":
    main()


import torch
from typing import Optional, Optional, Callable

from torch_geometric.data import Dataset, TemporalData
from tgb.linkproppred.dataset import LinkPropPredDataset
from tgb.linkproppred.negative_sampler import NegativeEdgeSampler


class PyGLinkPropPredDataset(Dataset):
    def __init__(
        self,
        name: str,
        root: str,
        transform: Optional[Callable] = None,
        pre_transform: Optional[Callable] = None,
    ):
        r"""
        PyG wrapper for the LinkPropPredDataset
        can return pytorch tensors for src,dst,t,msg,label
        can return Temporal Data object
        Parameters:
            name: name of the dataset, passed to `LinkPropPredDataset`
            root (string): Root directory where the dataset should be saved, passed to `LinkPropPredDataset`
            transform (callable, optional): A function/transform that takes in an, not used in this case
            pre_transform (callable, optional): A function/transform that takes in, not used in this case
        """
        self.name = name
        self.root = root
        self.dataset = LinkPropPredDataset(name=name, root=root)
        self._train_mask = torch.from_numpy(self.dataset.train_mask)
        self._val_mask = torch.from_numpy(self.dataset.val_mask)
        self._test_mask = torch.from_numpy(self.dataset.test_mask)
        super().__init__(root, transform, pre_transform)
        self._node_feat = self.dataset.node_feat
        self._edge_type = None
        self._static_data = None

        if self._node_feat is None:
            self._node_feat = None
        else:
            self._node_feat = torch.from_numpy(self._node_feat).float()
        
        self._node_type = self.dataset.node_type
        if self.node_type is not None:
            self._node_type = torch.from_numpy(self.dataset.node_type).long()
        
        self.process_data()

        self._ns_sampler = self.dataset.negative_sampler

    @property
    def eval_metric(self) -> str:
        """
        the official evaluation metric for the dataset, loaded from info.py
        Returns:
            eval_metric: str, the evaluation metric
        """
        return self.dataset.eval_metric

    @property
    def negative_sampler(self) -> NegativeEdgeSampler:
        r"""
        Returns the negative sampler of the dataset, will load negative samples from disc
        Returns:
            negative_sampler: NegativeEdgeSampler
        """
        return self._ns_sampler
    
    @property
    def num_nodes(self) -> int:
        r"""
        Returns the total number of unique nodes in the dataset 
        Returns:
            num_nodes: int, the number of unique nodes
        """
        return self.dataset.num_nodes
    
    @property
    def num_rels(self) -> int:
        r"""
        Returns the total number of unique relations in the dataset 
        Returns:
            num_rels: int, the number of unique relations
        """
        return self.dataset.num_rels
    
    @property
    def num_edges(self) -> int:
        r"""
        Returns the total number of edges in the dataset 
        Returns:
            num_edges: int, the number of edges
        """
        return self.dataset.num_edges

    def load_val_ns(self) -> None:
        r"""
        load the negative samples for the validation set
        """
        self.dataset.load_val_ns()

    def load_test_ns(self) -> None:
        r"""
        load the negative samples for the test set
        """
        self.dataset.load_test_ns()

    @property
    def train_mask(self) -> torch.Tensor:
        r"""
        Returns the train mask of the dataset
        Returns:
            train_mask: the mask for edges in the training set
        """
        if self._train_mask is None:
            raise ValueError("training split hasn't been loaded")
        return self._train_mask

    @property
    def val_mask(self) -> torch.Tensor:
        r"""
        Returns the validation mask of the dataset
        Returns:
            val_mask: the mask for edges in the validation set
        """
        if self._val_mask is None:
            raise ValueError("validation split hasn't been loaded")
        return self._val_mask

    @property
    def test_mask(self) -> torch.Tensor:
        r"""
        Returns the test mask of the dataset:
        Returns:
            test_mask: the mask for edges in the test set
        """
        if self._test_mask is None:
            raise ValueError("test split hasn't been loaded")
        return self._test_mask

    @property
    def node_feat(self) -> torch.Tensor:
        r"""
        Returns the node features of the dataset
        Returns:
            node_feat: the node features
        """
        return self._node_feat
    
    @property
    def node_type(self) -> torch.Tensor:
        r"""
        Returns the node types of the dataset
        Returns:
            node_type: the node types [N]
        """
        return self._node_type

    @property
    def src(self) -> torch.Tensor:
        r"""
        Returns the source nodes of the dataset
        Returns:
            src: the idx of the source nodes
        """
        return self._src

    @property
    def dst(self) -> torch.Tensor:
        r"""
        Returns the destination nodes of the dataset
        Returns:
            dst: the idx of the destination nodes
        """
        return self._dst

    @property
    def ts(self) -> torch.Tensor:
        r"""
        Returns the timestamps of the dataset
        Returns:
            ts: the timestamps of the edges
        """
        return self._ts
    
    @property
    def static_data(self) -> torch.Tensor:
        r"""
        Returns the static data of the dataset for tkgl-wikidata and tkgl-smallpedia
        Returns:
            static_data: the static data of the dataset
        """
        if (self._static_data is None):
            static_dict = {}
            static_dict["head"] = torch.from_numpy(self.dataset.static_data["head"]).long()
            static_dict["tail"] = torch.from_numpy(self.dataset.static_data["tail"]).long()
            static_dict["edge_type"] = torch.from_numpy(self.dataset.static_data["edge_type"]).long()
            self._static_data = static_dict
            return self._static_data
        else:
            return self._static_data 
    
    @property
    def edge_type(self) -> torch.Tensor:
        r"""
        Returns the edge types for each edge
        Returns:
            edge_type: edge type tensor (int)
        """
        return self._edge_type

    @property
    def edge_feat(self) -> torch.Tensor:
        r"""
        Returns the edge features of the dataset
        Returns:
            edge_feat: the edge features
        """
        return self._edge_feat

    @property
    def edge_label(self) -> torch.Tensor:
        r"""
        Returns the edge labels of the dataset
        Returns:
            edge_label: the labels of the edges
        """
        return self._edge_label

    def process_data(self) -> None:
        r"""
        convert the numpy arrays from dataset to pytorch tensors
        """
        src = torch.from_numpy(self.dataset.full_data["sources"])
        dst = torch.from_numpy(self.dataset.full_data["destinations"])
        ts = torch.from_numpy(self.dataset.full_data["timestamps"])
        msg = torch.from_numpy(
            self.dataset.full_data["edge_feat"]
        )  # use edge features here if available
        edge_label = torch.from_numpy(
            self.dataset.full_data["edge_label"]
        )  # this is the label indicating if an edge is a true edge, always 1 for true edges
        w = torch.from_numpy(
            self.dataset.full_data["w"]
        )


        # * first check typing for all tensors
        # source tensor must be of type int64
        # warnings.warn("sources tensor is not of type int64 or int32, forcing conversion")
        if src.dtype != torch.int64:
            src = src.long()

        # destination tensor must be of type int64
        if dst.dtype != torch.int64:
            dst = dst.long()

        # timestamp tensor must be of type int64
        if ts.dtype != torch.int64:
            ts = ts.long()

        # message tensor must be of type float32
        if msg.dtype != torch.float32:
            msg = msg.float()

        # weight tensor must be of type float32
        if w.dtype != torch.float32:
            w = w.float()

        #* for tkg
        if ("edge_type" in self.dataset.full_data):
            edge_type = torch.from_numpy(self.dataset.full_data["edge_type"])
            if edge_type.dtype != torch.int64:
                edge_type = edge_type.long()
            self._edge_type = edge_type

        self._src = src
        self._dst = dst
        self._ts = ts
        self._edge_label = edge_label
        self._edge_feat = msg
        self._w = w

    def get_TemporalData(self) -> TemporalData:
        """
        return the TemporalData object for the entire dataset
        """
        if (self._edge_type is not None):
            data = TemporalData(
                src=self._src,
                dst=self._dst,
                t=self._ts,
                msg=self._edge_feat,
                y=self._edge_label,
                edge_type=self._edge_type,
                w=self._w,
            )
        else:
            data = TemporalData(
                src=self._src,
                dst=self._dst,
                t=self._ts,
                msg=self._edge_feat,
                y=self._edge_label,
                w=self._w,
            )
        return data

    def len(self) -> int:
        """
        size of the dataset
        Returns:
            size: int
        """
        return self._src.shape[0]

    def get(self, idx: int) -> TemporalData:
        """
        construct temporal data object for a single edge
        Parameters:
            idx: index of the edge
        Returns:
            data: TemporalData object
        """
        if (self._edge_type is not None):
            data = TemporalData(
                src=self._src[idx],
                dst=self._dst[idx],
                t=self._ts[idx],
                msg=self._edge_feat[idx],
                y=self._edge_label[idx],
                edge_type=self._edge_type[idx]
            )
        else:
            data = TemporalData(
                src=self._src[idx],
                dst=self._dst[idx],
                t=self._ts[idx],
                msg=self._edge_feat[idx],
                y=self._edge_label[idx],
            )
        return data

    def __repr__(self) -> str:
        return f"{self.name.capitalize()}()"
